# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-function-docstring, missing-module-docstring

import unittest
from inspect import signature
from math import pi
import numpy as np
from scipy.linalg import expm
from ddt import data, ddt, unpack

from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister
from qiskit.exceptions import QiskitError
from qiskit.circuit.exceptions import CircuitError
from qiskit.circuit import Gate, ControlledGate
from qiskit.circuit.library import (
    U1Gate,
    U2Gate,
    U3Gate,
    CU1Gate,
    CU3Gate,
    XXMinusYYGate,
    XXPlusYYGate,
    RZGate,
    XGate,
    YGate,
    GlobalPhaseGate,
)
from qiskit.quantum_info import Pauli
from qiskit.quantum_info.operators.predicates import matrix_equal, is_unitary_matrix
from qiskit.utils.optionals import HAS_TWEEDLEDUM
from qiskit.quantum_info import Operator
from qiskit import transpile
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestStandard1Q(QiskitTestCase):
    """Standard Extension Test. Gates with a single Qubit"""

    def setUp(self):
        super().setUp()
        self.qr = QuantumRegister(3, "q")
        self.qr2 = QuantumRegister(3, "r")
        self.cr = ClassicalRegister(3, "c")
        self.circuit = QuantumCircuit(self.qr, self.qr2, self.cr)

    def test_barrier(self):
        self.circuit.barrier(self.qr[1])
        self.assertEqual(len(self.circuit), 1)
        self.assertEqual(self.circuit[0].operation.name, "barrier")
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_barrier_wires(self):
        self.circuit.barrier(1)
        self.assertEqual(len(self.circuit), 1)
        self.assertEqual(self.circuit[0].operation.name, "barrier")
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_barrier_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.barrier, self.cr[0])
        self.assertRaises(CircuitError, qc.barrier, self.cr)
        self.assertRaises(CircuitError, qc.barrier, (self.qr, "a"))
        self.assertRaises(CircuitError, qc.barrier, 0.0)

    def test_conditional_barrier_invalid(self):
        qc = self.circuit
        barrier = qc.barrier(self.qr)
        with self.assertWarns(DeprecationWarning):
            self.assertRaises(QiskitError, barrier.c_if, self.cr, 0)

    def test_barrier_reg(self):
        self.circuit.barrier(self.qr)
        self.assertEqual(len(self.circuit), 1)
        self.assertEqual(self.circuit[0].operation.name, "barrier")
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1], self.qr[2]))

    def test_barrier_none(self):
        self.circuit.barrier()
        self.assertEqual(len(self.circuit), 1)
        self.assertEqual(self.circuit[0].operation.name, "barrier")
        self.assertEqual(
            self.circuit[0].qubits,
            (self.qr[0], self.qr[1], self.qr[2], self.qr2[0], self.qr2[1], self.qr2[2]),
        )

    def test_ccx(self):
        self.circuit.ccx(self.qr[0], self.qr[1], self.qr[2])
        self.assertEqual(self.circuit[0].operation.name, "ccx")
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1], self.qr[2]))

    def test_ccx_wires(self):
        self.circuit.ccx(0, 1, 2)
        self.assertEqual(self.circuit[0].operation.name, "ccx")
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1], self.qr[2]))

    def test_ccx_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.ccx, self.cr[0], self.cr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.ccx, self.qr[0], self.qr[0], self.qr[2])
        self.assertRaises(CircuitError, qc.ccx, 0.0, self.qr[0], self.qr[2])
        self.assertRaises(CircuitError, qc.ccx, self.cr, self.qr, self.qr)
        self.assertRaises(CircuitError, qc.ccx, "a", self.qr[1], self.qr[2])

    def test_ch(self):
        self.circuit.ch(self.qr[0], self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "ch")
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1]))

    def test_ch_wires(self):
        self.circuit.ch(0, 1)
        self.assertEqual(self.circuit[0].operation.name, "ch")
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1]))

    def test_ch_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.ch, self.cr[0], self.cr[1])
        self.assertRaises(CircuitError, qc.ch, self.qr[0], self.qr[0])
        self.assertRaises(CircuitError, qc.ch, 0.0, self.qr[0])
        self.assertRaises(CircuitError, qc.ch, (self.qr, 3), self.qr[0])
        self.assertRaises(CircuitError, qc.ch, self.cr, self.qr)
        self.assertRaises(CircuitError, qc.ch, "a", self.qr[1])

    def test_cif_reg(self):
        with self.assertWarns(DeprecationWarning):
            self.circuit.h(self.qr[0]).c_if(self.cr, 7)
        self.assertEqual(self.circuit[0].operation.name, "h")
        self.assertEqual(self.circuit[0].qubits, (self.qr[0],))
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(self.circuit[0].operation.condition, (self.cr, 7))

    def test_cif_single_bit(self):
        with self.assertWarns(DeprecationWarning):
            self.circuit.h(self.qr[0]).c_if(self.cr[0], True)
        self.assertEqual(self.circuit[0].operation.name, "h")
        self.assertEqual(self.circuit[0].qubits, (self.qr[0],))
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(self.circuit[0].operation.condition, (self.cr[0], True))

    def test_crz(self):
        self.circuit.crz(1, self.qr[0], self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "crz")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1]))

    def test_cry(self):
        self.circuit.cry(1, self.qr[0], self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "cry")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1]))

    def test_crx(self):
        self.circuit.crx(1, self.qr[0], self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "crx")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1]))

    def test_crz_wires(self):
        self.circuit.crz(1, 0, 1)
        self.assertEqual(self.circuit[0].operation.name, "crz")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1]))

    def test_cry_wires(self):
        self.circuit.cry(1, 0, 1)
        self.assertEqual(self.circuit[0].operation.name, "cry")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1]))

    def test_crx_wires(self):
        self.circuit.crx(1, 0, 1)
        self.assertEqual(self.circuit[0].operation.name, "crx")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1]))

    def test_crz_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.crz, 0, self.cr[0], self.cr[1])
        self.assertRaises(CircuitError, qc.crz, 0, self.qr[0], self.qr[0])
        self.assertRaises(CircuitError, qc.crz, 0, 0.0, self.qr[0])
        self.assertRaises(CircuitError, qc.crz, self.qr[2], self.qr[1], self.qr[0])
        self.assertRaises(CircuitError, qc.crz, 0, self.qr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.crz, 0, (self.qr, 3), self.qr[1])
        self.assertRaises(CircuitError, qc.crz, 0, self.cr, self.qr)
        # TODO self.assertRaises(CircuitError, qc.crz, 'a', self.qr[1], self.qr[2])

    def test_cry_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.cry, 0, self.cr[0], self.cr[1])
        self.assertRaises(CircuitError, qc.cry, 0, self.qr[0], self.qr[0])
        self.assertRaises(CircuitError, qc.cry, 0, 0.0, self.qr[0])
        self.assertRaises(CircuitError, qc.cry, self.qr[2], self.qr[1], self.qr[0])
        self.assertRaises(CircuitError, qc.cry, 0, self.qr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.cry, 0, (self.qr, 3), self.qr[1])
        self.assertRaises(CircuitError, qc.cry, 0, self.cr, self.qr)
        # TODO self.assertRaises(CircuitError, qc.cry, 'a', self.qr[1], self.qr[2])

    def test_crx_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.crx, 0, self.cr[0], self.cr[1])
        self.assertRaises(CircuitError, qc.crx, 0, self.qr[0], self.qr[0])
        self.assertRaises(CircuitError, qc.crx, 0, 0.0, self.qr[0])
        self.assertRaises(CircuitError, qc.crx, self.qr[2], self.qr[1], self.qr[0])
        self.assertRaises(CircuitError, qc.crx, 0, self.qr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.crx, 0, (self.qr, 3), self.qr[1])
        self.assertRaises(CircuitError, qc.crx, 0, self.cr, self.qr)
        # TODO self.assertRaises(CircuitError, qc.crx, 'a', self.qr[1], self.qr[2])

    def test_cswap(self):
        self.circuit.cswap(self.qr[0], self.qr[1], self.qr[2])
        self.assertEqual(self.circuit[0].operation.name, "cswap")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1], self.qr[2]))

    def test_cswap_wires(self):
        self.circuit.cswap(0, 1, 2)
        self.assertEqual(self.circuit[0].operation.name, "cswap")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1], self.qr[2]))

    def test_cswap_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.cswap, self.cr[0], self.cr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.cswap, self.qr[1], self.qr[0], self.qr[0])
        self.assertRaises(CircuitError, qc.cswap, self.qr[1], 0.0, self.qr[0])
        self.assertRaises(CircuitError, qc.cswap, self.cr[0], self.cr[1], self.qr[0])
        self.assertRaises(CircuitError, qc.cswap, self.qr[0], self.qr[0], self.qr[1])
        self.assertRaises(CircuitError, qc.cswap, 0.0, self.qr[0], self.qr[1])
        self.assertRaises(CircuitError, qc.cswap, (self.qr, 3), self.qr[0], self.qr[1])
        self.assertRaises(CircuitError, qc.cswap, self.cr, self.qr[0], self.qr[1])
        self.assertRaises(CircuitError, qc.cswap, "a", self.qr[1], self.qr[2])

    def test_cu1(self):
        self.circuit.append(CU1Gate(1), [self.qr[1], self.qr[2]])
        self.assertEqual(self.circuit[0].operation.name, "cu1")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_cu1_wires(self):
        self.circuit.append(CU1Gate(1), [1, 2])
        self.assertEqual(self.circuit[0].operation.name, "cu1")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_cu3(self):
        self.circuit.append(CU3Gate(1, 2, 3), [self.qr[1], self.qr[2]])
        self.assertEqual(self.circuit[0].operation.name, "cu3")
        self.assertEqual(self.circuit[0].operation.params, [1, 2, 3])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_cu3_wires(self):
        self.circuit.append(CU3Gate(1, 2, 3), [1, 2])
        self.assertEqual(self.circuit[0].operation.name, "cu3")
        self.assertEqual(self.circuit[0].operation.params, [1, 2, 3])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_cx(self):
        self.circuit.cx(self.qr[1], self.qr[2])
        self.assertEqual(self.circuit[0].operation.name, "cx")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_cx_wires(self):
        self.circuit.cx(1, 2)
        self.assertEqual(self.circuit[0].operation.name, "cx")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_cx_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.cx, self.cr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.cx, self.qr[0], self.qr[0])
        self.assertRaises(CircuitError, qc.cx, 0.0, self.qr[0])
        self.assertRaises(CircuitError, qc.cx, (self.qr, 3), self.qr[0])
        self.assertRaises(CircuitError, qc.cx, self.cr, self.qr)
        self.assertRaises(CircuitError, qc.cx, "a", self.qr[1])

    def test_cy(self):
        self.circuit.cy(self.qr[1], self.qr[2])
        self.assertEqual(self.circuit[0].operation.name, "cy")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_cy_wires(self):
        self.circuit.cy(1, 2)
        self.assertEqual(self.circuit[0].operation.name, "cy")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_cy_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.cy, self.cr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.cy, self.qr[0], self.qr[0])
        self.assertRaises(CircuitError, qc.cy, 0.0, self.qr[0])
        self.assertRaises(CircuitError, qc.cy, (self.qr, 3), self.qr[0])
        self.assertRaises(CircuitError, qc.cy, self.cr, self.qr)
        self.assertRaises(CircuitError, qc.cy, "a", self.qr[1])

    def test_cz(self):
        self.circuit.cz(self.qr[1], self.qr[2])
        self.assertEqual(self.circuit[0].operation.name, "cz")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_cz_wires(self):
        self.circuit.cz(1, 2)
        self.assertEqual(self.circuit[0].operation.name, "cz")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_cz_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.cz, self.cr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.cz, self.qr[0], self.qr[0])
        self.assertRaises(CircuitError, qc.cz, 0.0, self.qr[0])
        self.assertRaises(CircuitError, qc.cz, (self.qr, 3), self.qr[0])
        self.assertRaises(CircuitError, qc.cz, self.cr, self.qr)
        self.assertRaises(CircuitError, qc.cz, "a", self.qr[1])

    def test_h(self):
        self.circuit.h(self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "h")
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_h_wires(self):
        self.circuit.h(1)
        self.assertEqual(self.circuit[0].operation.name, "h")
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_h_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.h, self.cr[0])
        self.assertRaises(CircuitError, qc.h, self.cr)
        self.assertRaises(CircuitError, qc.h, (self.qr, 3))
        self.assertRaises(CircuitError, qc.h, (self.qr, "a"))
        self.assertRaises(CircuitError, qc.h, 0.0)

    def test_h_reg(self):
        instruction_set = self.circuit.h(self.qr)
        self.assertEqual(len(instruction_set), 3)
        self.assertEqual(instruction_set[0].operation.name, "h")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))

    def test_h_reg_inv(self):
        instruction_set = self.circuit.h(self.qr).inverse()
        self.assertEqual(len(instruction_set), 3)
        self.assertEqual(instruction_set[0].operation.name, "h")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))

    def test_iden(self):
        self.circuit.id(self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "id")
        self.assertEqual(self.circuit[0].operation.params, [])

    def test_iden_wires(self):
        self.circuit.id(1)
        self.assertEqual(self.circuit[0].operation.name, "id")
        self.assertEqual(self.circuit[0].operation.params, [])

    def test_iden_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.id, self.cr[0])
        self.assertRaises(CircuitError, qc.id, self.cr)
        self.assertRaises(CircuitError, qc.id, (self.qr, 3))
        self.assertRaises(CircuitError, qc.id, (self.qr, "a"))
        self.assertRaises(CircuitError, qc.id, 0.0)

    def test_iden_reg(self):
        instruction_set = self.circuit.id(self.qr)
        self.assertEqual(len(instruction_set), 3)
        self.assertEqual(instruction_set[0].operation.name, "id")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))

    def test_iden_reg_inv(self):
        instruction_set = self.circuit.id(self.qr).inverse()
        self.assertEqual(len(instruction_set), 3)
        self.assertEqual(instruction_set[0].operation.name, "id")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))

    def test_rx(self):
        self.circuit.rx(1, self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "rx")
        self.assertEqual(self.circuit[0].operation.params, [1])

    def test_rx_wires(self):
        self.circuit.rx(1, 1)
        self.assertEqual(self.circuit[0].operation.name, "rx")
        self.assertEqual(self.circuit[0].operation.params, [1])

    def test_rx_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.rx, self.cr[0], self.cr[1])
        self.assertRaises(CircuitError, qc.rx, self.qr[1], 0)
        self.assertRaises(CircuitError, qc.rx, 0, self.cr[0])
        self.assertRaises(CircuitError, qc.rx, 0, 0.0)
        self.assertRaises(CircuitError, qc.rx, self.qr[2], self.qr[1])
        self.assertRaises(CircuitError, qc.rx, 0, (self.qr, 3))
        self.assertRaises(CircuitError, qc.rx, 0, self.cr)
        # TODO self.assertRaises(CircuitError, qc.rx, 'a', self.qr[1])
        self.assertRaises(CircuitError, qc.rx, 0, "a")

    def test_rx_reg(self):
        instruction_set = self.circuit.rx(1, self.qr)
        self.assertEqual(len(instruction_set), 3)
        self.assertEqual(instruction_set[0].operation.name, "rx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_rx_reg_inv(self):
        instruction_set = self.circuit.rx(1, self.qr).inverse()
        self.assertEqual(len(instruction_set), 3)
        self.assertEqual(instruction_set[0].operation.name, "rx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_rx_pi(self):
        qc = self.circuit
        qc.rx(pi / 2, self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "rx")
        self.assertEqual(self.circuit[0].operation.params, [pi / 2])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_ry(self):
        self.circuit.ry(1, self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "ry")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_ry_wires(self):
        self.circuit.ry(1, 1)
        self.assertEqual(self.circuit[0].operation.name, "ry")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_ry_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.ry, self.cr[0], self.cr[1])
        self.assertRaises(CircuitError, qc.ry, self.qr[1], 0)
        self.assertRaises(CircuitError, qc.ry, 0, self.cr[0])
        self.assertRaises(CircuitError, qc.ry, 0, 0.0)
        self.assertRaises(CircuitError, qc.ry, self.qr[2], self.qr[1])
        self.assertRaises(CircuitError, qc.ry, 0, (self.qr, 3))
        self.assertRaises(CircuitError, qc.ry, 0, self.cr)
        # TODO self.assertRaises(CircuitError, qc.ry, 'a', self.qr[1])
        self.assertRaises(CircuitError, qc.ry, 0, "a")

    def test_ry_reg(self):
        instruction_set = self.circuit.ry(1, self.qr)
        self.assertEqual(instruction_set[0].operation.name, "ry")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_ry_reg_inv(self):
        instruction_set = self.circuit.ry(1, self.qr).inverse()
        self.assertEqual(instruction_set[0].operation.name, "ry")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_ry_pi(self):
        qc = self.circuit
        qc.ry(pi / 2, self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "ry")
        self.assertEqual(self.circuit[0].operation.params, [pi / 2])

    def test_rz(self):
        self.circuit.rz(1, self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "rz")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_rz_wires(self):
        self.circuit.rz(1, 1)
        self.assertEqual(self.circuit[0].operation.name, "rz")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_rz_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.rz, self.cr[0], self.cr[1])
        self.assertRaises(CircuitError, qc.rz, self.qr[1], 0)
        self.assertRaises(CircuitError, qc.rz, 0, self.cr[0])
        self.assertRaises(CircuitError, qc.rz, 0, 0.0)
        self.assertRaises(CircuitError, qc.rz, self.qr[2], self.qr[1])
        self.assertRaises(CircuitError, qc.rz, 0, (self.qr, 3))
        self.assertRaises(CircuitError, qc.rz, 0, self.cr)
        # TODO self.assertRaises(CircuitError, qc.rz, 'a', self.qr[1])
        self.assertRaises(CircuitError, qc.rz, 0, "a")

    def test_rz_reg(self):
        instruction_set = self.circuit.rz(1, self.qr)
        self.assertEqual(instruction_set[0].operation.name, "rz")
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_rz_reg_inv(self):
        instruction_set = self.circuit.rz(1, self.qr).inverse()
        self.assertEqual(instruction_set[0].operation.name, "rz")
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_rz_pi(self):
        self.circuit.rz(pi / 2, self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "rz")
        self.assertEqual(self.circuit[0].operation.params, [pi / 2])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_rzz(self):
        self.circuit.rzz(1, self.qr[1], self.qr[2])
        self.assertEqual(self.circuit[0].operation.name, "rzz")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_rzz_wires(self):
        self.circuit.rzz(1, 1, 2)
        self.assertEqual(self.circuit[0].operation.name, "rzz")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_rzz_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.rzz, 1, self.cr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.rzz, 1, self.qr[0], self.qr[0])
        self.assertRaises(CircuitError, qc.rzz, 1, 0.0, self.qr[0])
        self.assertRaises(CircuitError, qc.rzz, 1, (self.qr, 3), self.qr[0])
        self.assertRaises(CircuitError, qc.rzz, 1, self.cr, self.qr)
        self.assertRaises(CircuitError, qc.rzz, 1, "a", self.qr[1])
        self.assertRaises(CircuitError, qc.rzz, 0.1, self.cr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.rzz, 0.1, self.qr[0], self.qr[0])

    def test_s(self):
        self.circuit.s(self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "s")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_s_wires(self):
        self.circuit.s(1)
        self.assertEqual(self.circuit[0].operation.name, "s")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_s_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.s, self.cr[0])
        self.assertRaises(CircuitError, qc.s, self.cr)
        self.assertRaises(CircuitError, qc.s, (self.qr, 3))
        self.assertRaises(CircuitError, qc.s, (self.qr, "a"))
        self.assertRaises(CircuitError, qc.s, 0.0)

    def test_s_reg(self):
        instruction_set = self.circuit.s(self.qr)
        self.assertEqual(instruction_set[0].operation.name, "s")
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_s_reg_inv(self):
        instruction_set = self.circuit.s(self.qr).inverse()
        self.assertEqual(instruction_set[0].operation.name, "sdg")
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_sdg(self):
        self.circuit.sdg(self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "sdg")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_sdg_wires(self):
        self.circuit.sdg(1)
        self.assertEqual(self.circuit[0].operation.name, "sdg")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_sdg_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.sdg, self.cr[0])
        self.assertRaises(CircuitError, qc.sdg, self.cr)
        self.assertRaises(CircuitError, qc.sdg, (self.qr, 3))
        self.assertRaises(CircuitError, qc.sdg, (self.qr, "a"))
        self.assertRaises(CircuitError, qc.sdg, 0.0)

    def test_sdg_reg(self):
        instruction_set = self.circuit.sdg(self.qr)
        self.assertEqual(instruction_set[0].operation.name, "sdg")
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_sdg_reg_inv(self):
        instruction_set = self.circuit.sdg(self.qr).inverse()
        self.assertEqual(instruction_set[0].operation.name, "s")
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_swap(self):
        self.circuit.swap(self.qr[1], self.qr[2])
        self.assertEqual(self.circuit[0].operation.name, "swap")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_swap_wires(self):
        self.circuit.swap(1, 2)
        self.assertEqual(self.circuit[0].operation.name, "swap")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1], self.qr[2]))

    def test_swap_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.swap, self.cr[1], self.cr[2])
        self.assertRaises(CircuitError, qc.swap, self.qr[0], self.qr[0])
        self.assertRaises(CircuitError, qc.swap, 0.0, self.qr[0])
        self.assertRaises(CircuitError, qc.swap, (self.qr, 3), self.qr[0])
        self.assertRaises(CircuitError, qc.swap, self.cr, self.qr)
        self.assertRaises(CircuitError, qc.swap, "a", self.qr[1])
        self.assertRaises(CircuitError, qc.swap, self.qr, self.qr2[[1, 2]])
        self.assertRaises(CircuitError, qc.swap, self.qr[:2], self.qr2)

    def test_t(self):
        self.circuit.t(self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "t")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_t_wire(self):
        self.circuit.t(1)
        self.assertEqual(self.circuit[0].operation.name, "t")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_t_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.t, self.cr[0])
        self.assertRaises(CircuitError, qc.t, self.cr)
        self.assertRaises(CircuitError, qc.t, (self.qr, 3))
        self.assertRaises(CircuitError, qc.t, (self.qr, "a"))
        self.assertRaises(CircuitError, qc.t, 0.0)

    def test_t_reg(self):
        instruction_set = self.circuit.t(self.qr)
        self.assertEqual(instruction_set[0].operation.name, "t")
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_t_reg_inv(self):
        instruction_set = self.circuit.t(self.qr).inverse()
        self.assertEqual(instruction_set[0].operation.name, "tdg")
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_tdg(self):
        self.circuit.tdg(self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "tdg")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_tdg_wires(self):
        self.circuit.tdg(1)
        self.assertEqual(self.circuit[0].operation.name, "tdg")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_tdg_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.tdg, self.cr[0])
        self.assertRaises(CircuitError, qc.tdg, self.cr)
        self.assertRaises(CircuitError, qc.tdg, (self.qr, 3))
        self.assertRaises(CircuitError, qc.tdg, (self.qr, "a"))
        self.assertRaises(CircuitError, qc.tdg, 0.0)

    def test_tdg_reg(self):
        instruction_set = self.circuit.tdg(self.qr)
        self.assertEqual(instruction_set[0].operation.name, "tdg")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_tdg_reg_inv(self):
        instruction_set = self.circuit.tdg(self.qr).inverse()
        self.assertEqual(instruction_set[0].operation.name, "t")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_u1(self):
        self.circuit.append(U1Gate(1), [self.qr[1]])
        self.assertEqual(self.circuit[0].operation.name, "u1")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_u1_wires(self):
        self.circuit.append(U1Gate(1), [1])
        self.assertEqual(self.circuit[0].operation.name, "u1")
        self.assertEqual(self.circuit[0].operation.params, [1])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_u1_reg(self):
        instruction_set = self.circuit.append(U1Gate(1), [self.qr])
        self.assertEqual(instruction_set[0].operation.name, "u1")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_u1_reg_inv(self):
        instruction_set = self.circuit.append(U1Gate(1), [self.qr]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "u1")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_u1_pi(self):
        qc = self.circuit
        qc.append(U1Gate(pi / 2), [self.qr[1]])
        self.assertEqual(self.circuit[0].operation.name, "u1")
        self.assertEqual(self.circuit[0].operation.params, [pi / 2])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_u2(self):
        self.circuit.append(U2Gate(1, 2), [self.qr[1]])
        self.assertEqual(self.circuit[0].operation.name, "u2")
        self.assertEqual(self.circuit[0].operation.params, [1, 2])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_u2_wires(self):
        self.circuit.append(U2Gate(1, 2), [1])
        self.assertEqual(self.circuit[0].operation.name, "u2")
        self.assertEqual(self.circuit[0].operation.params, [1, 2])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_u2_reg(self):
        instruction_set = self.circuit.append(U2Gate(1, 2), [self.qr])
        self.assertEqual(instruction_set[0].operation.name, "u2")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [1, 2])

    def test_u2_reg_inv(self):
        instruction_set = self.circuit.append(U2Gate(1, 2), [self.qr]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "u2")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [-pi - 2, -1 + pi])

    def test_u2_pi(self):
        self.circuit.append(U2Gate(pi / 2, 0.3 * pi), [self.qr[1]])
        self.assertEqual(self.circuit[0].operation.name, "u2")
        self.assertEqual(self.circuit[0].operation.params, [pi / 2, 0.3 * pi])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_u3(self):
        self.circuit.append(U3Gate(1, 2, 3), [self.qr[1]])
        self.assertEqual(self.circuit[0].operation.name, "u3")
        self.assertEqual(self.circuit[0].operation.params, [1, 2, 3])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_u3_wires(self):
        self.circuit.append(U3Gate(1, 2, 3), [1])
        self.assertEqual(self.circuit[0].operation.name, "u3")
        self.assertEqual(self.circuit[0].operation.params, [1, 2, 3])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_u3_reg(self):
        instruction_set = self.circuit.append(U3Gate(1, 2, 3), [self.qr])
        self.assertEqual(instruction_set[0].operation.name, "u3")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [1, 2, 3])

    def test_u3_reg_inv(self):
        instruction_set = self.circuit.append(U3Gate(1, 2, 3), [self.qr]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "u3")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [-1, -3, -2])

    def test_u3_pi(self):
        self.circuit.append(U3Gate(pi, pi / 2, 0.3 * pi), [self.qr[1]])
        self.assertEqual(self.circuit[0].operation.name, "u3")
        self.assertEqual(self.circuit[0].operation.params, [pi, pi / 2, 0.3 * pi])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_x(self):
        self.circuit.x(self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "x")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_x_wires(self):
        self.circuit.x(1)
        self.assertEqual(self.circuit[0].operation.name, "x")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_x_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.x, self.cr[0])
        self.assertRaises(CircuitError, qc.x, self.cr)
        self.assertRaises(CircuitError, qc.x, (self.qr, "a"))
        self.assertRaises(CircuitError, qc.x, 0.0)

    def test_x_reg(self):
        instruction_set = self.circuit.x(self.qr)
        self.assertEqual(instruction_set[0].operation.name, "x")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_x_reg_inv(self):
        instruction_set = self.circuit.x(self.qr).inverse()
        self.assertEqual(instruction_set[0].operation.name, "x")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_y(self):
        self.circuit.y(self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "y")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_y_wires(self):
        self.circuit.y(1)
        self.assertEqual(self.circuit[0].operation.name, "y")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_y_invalid(self):
        qc = self.circuit
        self.assertRaises(CircuitError, qc.y, self.cr[0])
        self.assertRaises(CircuitError, qc.y, self.cr)
        self.assertRaises(CircuitError, qc.y, (self.qr, "a"))
        self.assertRaises(CircuitError, qc.y, 0.0)

    def test_y_reg(self):
        instruction_set = self.circuit.y(self.qr)
        self.assertEqual(instruction_set[0].operation.name, "y")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_y_reg_inv(self):
        instruction_set = self.circuit.y(self.qr).inverse()
        self.assertEqual(instruction_set[0].operation.name, "y")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_z(self):
        self.circuit.z(self.qr[1])
        self.assertEqual(self.circuit[0].operation.name, "z")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_z_wires(self):
        self.circuit.z(1)
        self.assertEqual(self.circuit[0].operation.name, "z")
        self.assertEqual(self.circuit[0].operation.params, [])
        self.assertEqual(self.circuit[0].qubits, (self.qr[1],))

    def test_z_reg(self):
        instruction_set = self.circuit.z(self.qr)
        self.assertEqual(instruction_set[0].operation.name, "z")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_z_reg_inv(self):
        instruction_set = self.circuit.z(self.qr).inverse()
        self.assertEqual(instruction_set[0].operation.name, "z")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1],))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_global_phase(self):
        qc = self.circuit
        qc.append(GlobalPhaseGate(0.1), [])
        self.assertEqual(self.circuit[0].operation.name, "global_phase")
        self.assertEqual(self.circuit[0].operation.params, [0.1])
        self.assertEqual(self.circuit[0].qubits, ())

    def test_global_phase_inv(self):
        instruction_set = self.circuit.append(GlobalPhaseGate(0.1), []).inverse()
        self.assertEqual(len(instruction_set), 1)
        self.assertEqual(instruction_set[0].operation.params, [-0.1])

    def test_global_phase_matrix(self):
        """Test global_phase matrix."""
        theta = 0.1
        np.testing.assert_allclose(
            np.array(GlobalPhaseGate(theta)),
            np.array([[np.exp(1j * theta)]], dtype=complex),
            atol=1e-7,
        )

    def test_global_phase_consistency(self):
        """Tests compatibility of GlobalPhaseGate with QuantumCircuit.global_phase"""
        theta = 0.1
        qc1 = QuantumCircuit(0, global_phase=theta)
        qc2 = QuantumCircuit(0)
        qc2.append(GlobalPhaseGate(theta), [])
        np.testing.assert_allclose(
            Operator(qc1),
            Operator(qc2),
            atol=1e-7,
        )

    def test_transpile_global_phase_consistency(self):
        """Tests compatibility of transpiled GlobalPhaseGate with QuantumCircuit.global_phase"""
        qc1 = QuantumCircuit(0, global_phase=0.3)
        qc2 = QuantumCircuit(0, global_phase=0.2)
        qc2.append(GlobalPhaseGate(0.1), [])
        np.testing.assert_allclose(
            Operator(transpile(qc1, basis_gates=["u"])),
            Operator(transpile(qc2, basis_gates=["u"])),
            atol=1e-7,
        )


@ddt
class TestStandard2Q(QiskitTestCase):
    """Standard Extension Test. Gates with two Qubits"""

    def setUp(self):
        super().setUp()
        self.qr = QuantumRegister(3, "q")
        self.qr2 = QuantumRegister(3, "r")
        self.cr = ClassicalRegister(3, "c")
        self.circuit = QuantumCircuit(self.qr, self.qr2, self.cr)

    def test_barrier_reg_bit(self):
        self.circuit.barrier(self.qr, self.qr2[0])
        self.assertEqual(len(self.circuit), 1)
        self.assertEqual(self.circuit[0].operation.name, "barrier")
        self.assertEqual(self.circuit[0].qubits, (self.qr[0], self.qr[1], self.qr[2], self.qr2[0]))

    def test_ch_reg_reg(self):
        instruction_set = self.circuit.ch(self.qr, self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "ch")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_ch_reg_reg_inv(self):
        instruction_set = self.circuit.ch(self.qr, self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "ch")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_ch_reg_bit(self):
        instruction_set = self.circuit.ch(self.qr, self.qr2[1])
        self.assertEqual(instruction_set[0].operation.name, "ch")
        self.assertEqual(
            instruction_set[1].qubits,
            (
                self.qr[1],
                self.qr2[1],
            ),
        )
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_ch_reg_bit_inv(self):
        instruction_set = self.circuit.ch(self.qr, self.qr2[1]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "ch")
        self.assertEqual(
            instruction_set[1].qubits,
            (
                self.qr[1],
                self.qr2[1],
            ),
        )
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_ch_bit_reg(self):
        instruction_set = self.circuit.ch(self.qr[1], self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "ch")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_crz_reg_reg(self):
        instruction_set = self.circuit.crz(1, self.qr, self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "crz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_crz_reg_reg_inv(self):
        instruction_set = self.circuit.crz(1, self.qr, self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "crz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_crz_reg_bit(self):
        instruction_set = self.circuit.crz(1, self.qr, self.qr2[1])
        self.assertEqual(instruction_set[0].operation.name, "crz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_crz_reg_bit_inv(self):
        instruction_set = self.circuit.crz(1, self.qr, self.qr2[1]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "crz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_crz_bit_reg(self):
        instruction_set = self.circuit.crz(1, self.qr[1], self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "crz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_crz_bit_reg_inv(self):
        instruction_set = self.circuit.crz(1, self.qr[1], self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "crz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_cry_reg_reg(self):
        instruction_set = self.circuit.cry(1, self.qr, self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "cry")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_cry_reg_reg_inv(self):
        instruction_set = self.circuit.cry(1, self.qr, self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cry")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_cry_reg_bit(self):
        instruction_set = self.circuit.cry(1, self.qr, self.qr2[1])
        self.assertEqual(instruction_set[0].operation.name, "cry")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_cry_reg_bit_inv(self):
        instruction_set = self.circuit.cry(1, self.qr, self.qr2[1]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cry")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_cry_bit_reg(self):
        instruction_set = self.circuit.cry(1, self.qr[1], self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "cry")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_cry_bit_reg_inv(self):
        instruction_set = self.circuit.cry(1, self.qr[1], self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cry")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_crx_reg_reg(self):
        instruction_set = self.circuit.crx(1, self.qr, self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "crx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_crx_reg_reg_inv(self):
        instruction_set = self.circuit.crx(1, self.qr, self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "crx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_crx_reg_bit(self):
        instruction_set = self.circuit.crx(1, self.qr, self.qr2[1])
        self.assertEqual(instruction_set[0].operation.name, "crx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_crx_reg_bit_inv(self):
        instruction_set = self.circuit.crx(1, self.qr, self.qr2[1]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "crx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_crx_bit_reg(self):
        instruction_set = self.circuit.crx(1, self.qr[1], self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "crx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_crx_bit_reg_inv(self):
        instruction_set = self.circuit.crx(1, self.qr[1], self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "crx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_cu1_reg_reg(self):
        instruction_set = self.circuit.append(CU1Gate(1), [self.qr, self.qr2])
        self.assertEqual(instruction_set[0].operation.name, "cu1")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_cu1_reg_reg_inv(self):
        instruction_set = self.circuit.append(CU1Gate(1), [self.qr, self.qr2]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cu1")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_cu1_reg_bit(self):
        instruction_set = self.circuit.append(CU1Gate(1), [self.qr, self.qr2[1]])
        self.assertEqual(instruction_set[0].operation.name, "cu1")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_cu1_reg_bit_inv(self):
        instruction_set = self.circuit.append(CU1Gate(1), [self.qr, self.qr2[1]]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cu1")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_cu1_bit_reg(self):
        instruction_set = self.circuit.append(CU1Gate(1), [self.qr[1], self.qr2])
        self.assertEqual(instruction_set[0].operation.name, "cu1")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1])

    def test_cu1_bit_reg_inv(self):
        instruction_set = self.circuit.append(CU1Gate(1), [self.qr[1], self.qr2]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cu1")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1])

    def test_cu3_reg_reg(self):
        instruction_set = self.circuit.append(CU3Gate(1, 2, 3), [self.qr, self.qr2])
        self.assertEqual(instruction_set[0].operation.name, "cu3")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1, 2, 3])

    def test_cu3_reg_reg_inv(self):
        instruction_set = self.circuit.append(CU3Gate(1, 2, 3), [self.qr, self.qr2]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cu3")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1, -3, -2])

    def test_cu3_reg_bit(self):
        instruction_set = self.circuit.append(CU3Gate(1, 2, 3), [self.qr, self.qr2[1]])
        self.assertEqual(instruction_set[0].operation.name, "cu3")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1, 2, 3])

    def test_cu3_reg_bit_inv(self):
        instruction_set = self.circuit.append(CU3Gate(1, 2, 3), [self.qr, self.qr2[1]]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cu3")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1, -3, -2])

    def test_cu3_bit_reg(self):
        instruction_set = self.circuit.append(CU3Gate(1, 2, 3), [self.qr[1], self.qr2])
        self.assertEqual(instruction_set[0].operation.name, "cu3")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [1, 2, 3])

    def test_cu3_bit_reg_inv(self):
        instruction_set = self.circuit.append(CU3Gate(1, 2, 3), [self.qr[1], self.qr2]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cu3")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [-1, -3, -2])

    def test_cx_reg_reg(self):
        instruction_set = self.circuit.cx(self.qr, self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "cx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cx_reg_reg_inv(self):
        instruction_set = self.circuit.cx(self.qr, self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cx_reg_bit(self):
        instruction_set = self.circuit.cx(self.qr, self.qr2[1])
        self.assertEqual(instruction_set[0].operation.name, "cx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cx_reg_bit_inv(self):
        instruction_set = self.circuit.cx(self.qr, self.qr2[1]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cx_bit_reg(self):
        instruction_set = self.circuit.cx(self.qr[1], self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "cx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cx_bit_reg_inv(self):
        instruction_set = self.circuit.cx(self.qr[1], self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cy_reg_reg(self):
        instruction_set = self.circuit.cy(self.qr, self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "cy")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cy_reg_reg_inv(self):
        instruction_set = self.circuit.cy(self.qr, self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cy")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cy_reg_bit(self):
        instruction_set = self.circuit.cy(self.qr, self.qr2[1])
        self.assertEqual(instruction_set[0].operation.name, "cy")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cy_reg_bit_inv(self):
        instruction_set = self.circuit.cy(self.qr, self.qr2[1]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cy")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cy_bit_reg(self):
        instruction_set = self.circuit.cy(self.qr[1], self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "cy")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cy_bit_reg_inv(self):
        instruction_set = self.circuit.cy(self.qr[1], self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cy")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cz_reg_reg(self):
        instruction_set = self.circuit.cz(self.qr, self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "cz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cz_reg_reg_inv(self):
        instruction_set = self.circuit.cz(self.qr, self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cz_reg_bit(self):
        instruction_set = self.circuit.cz(self.qr, self.qr2[1])
        self.assertEqual(instruction_set[0].operation.name, "cz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cz_reg_bit_inv(self):
        instruction_set = self.circuit.cz(self.qr, self.qr2[1]).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cz_bit_reg(self):
        instruction_set = self.circuit.cz(self.qr[1], self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "cz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cz_bit_reg_inv(self):
        instruction_set = self.circuit.cz(self.qr[1], self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cz")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_swap_reg_reg(self):
        instruction_set = self.circuit.swap(self.qr, self.qr2)
        self.assertEqual(instruction_set[0].operation.name, "swap")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_swap_reg_reg_inv(self):
        instruction_set = self.circuit.swap(self.qr, self.qr2).inverse()
        self.assertEqual(instruction_set[0].operation.name, "swap")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    @unpack
    @data(
        (0, 0, np.eye(4)),
        (
            np.pi / 2,
            np.pi / 2,
            np.array(
                [
                    [np.sqrt(2) / 2, 0, 0, -np.sqrt(2) / 2],
                    [0, 1, 0, 0],
                    [0, 0, 1, 0],
                    [np.sqrt(2) / 2, 0, 0, np.sqrt(2) / 2],
                ]
            ),
        ),
        (
            np.pi,
            np.pi / 2,
            np.array([[0, 0, 0, -1], [0, 1, 0, 0], [0, 0, 1, 0], [1, 0, 0, 0]]),
        ),
        (
            2 * np.pi,
            np.pi / 2,
            np.array([[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]]),
        ),
        (
            np.pi / 2,
            np.pi,
            np.array(
                [
                    [np.sqrt(2) / 2, 0, 0, 1j * np.sqrt(2) / 2],
                    [0, 1, 0, 0],
                    [0, 0, 1, 0],
                    [1j * np.sqrt(2) / 2, 0, 0, np.sqrt(2) / 2],
                ]
            ),
        ),
        (4 * np.pi, 0, np.eye(4)),
    )
    def test_xx_minus_yy_matrix(self, theta: float, beta: float, expected: np.ndarray):
        """Test XX-YY matrix."""
        gate = XXMinusYYGate(theta, beta)
        np.testing.assert_allclose(np.array(gate), expected, atol=1e-7)

    def test_xx_minus_yy_exponential_formula(self):
        """Test XX-YY exponential formula."""
        theta, beta = np.random.uniform(-10, 10, size=2)
        gate = XXMinusYYGate(theta, beta)
        x = np.array(XGate())
        y = np.array(YGate())
        xx = np.kron(x, x)
        yy = np.kron(y, y)
        rz1 = np.kron(np.array(RZGate(beta)), np.eye(2))
        np.testing.assert_allclose(
            np.array(gate),
            rz1 @ expm(-0.25j * theta * (xx - yy)) @ rz1.T.conj(),
            atol=1e-7,
        )

    def test_xx_plus_yy_exponential_formula(self):
        """Test XX+YY exponential formula."""
        theta, beta = np.random.uniform(-10, 10, size=2)
        gate = XXPlusYYGate(theta, beta)
        x = np.array(XGate())
        y = np.array(YGate())
        xx = np.kron(x, x)
        yy = np.kron(y, y)
        rz0 = np.kron(np.eye(2), np.array(RZGate(beta)))
        np.testing.assert_allclose(
            np.array(gate),
            rz0.T.conj() @ expm(-0.25j * theta * (xx + yy)) @ rz0,
            atol=1e-7,
        )


class TestStandard3Q(QiskitTestCase):
    """Standard Extension Test. Gates with three Qubits"""

    def setUp(self):
        super().setUp()
        self.qr = QuantumRegister(3, "q")
        self.qr2 = QuantumRegister(3, "r")
        self.qr3 = QuantumRegister(3, "s")
        self.cr = ClassicalRegister(3, "c")
        self.circuit = QuantumCircuit(self.qr, self.qr2, self.qr3, self.cr)

    def test_ccx_reg_reg_reg(self):
        instruction_set = self.circuit.ccx(self.qr, self.qr2, self.qr3)
        self.assertEqual(instruction_set[0].operation.name, "ccx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1], self.qr3[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_ccx_reg_reg_inv(self):
        instruction_set = self.circuit.ccx(self.qr, self.qr2, self.qr3).inverse()
        self.assertEqual(instruction_set[0].operation.name, "ccx")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1], self.qr3[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cswap_reg_reg_reg(self):
        instruction_set = self.circuit.cswap(self.qr, self.qr2, self.qr3)
        self.assertEqual(instruction_set[0].operation.name, "cswap")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1], self.qr3[1]))
        self.assertEqual(instruction_set[2].operation.params, [])

    def test_cswap_reg_reg_inv(self):
        instruction_set = self.circuit.cswap(self.qr, self.qr2, self.qr3).inverse()
        self.assertEqual(instruction_set[0].operation.name, "cswap")
        self.assertEqual(instruction_set[1].qubits, (self.qr[1], self.qr2[1], self.qr3[1]))
        self.assertEqual(instruction_set[2].operation.params, [])


class TestStandardMethods(QiskitTestCase):
    """Standard Extension Test."""

    @unittest.skipUnless(HAS_TWEEDLEDUM, "tweedledum required for this test")
    def test_to_matrix(self):
        """test gates implementing to_matrix generate matrix which matches definition."""
        from qiskit.circuit.library.pauli_evolution import PauliEvolutionGate
        from qiskit.circuit.library.generalized_gates.pauli import PauliGate
        from qiskit.circuit.classicalfunction.boolean_expression import BooleanExpression

        params = [0.1 * (i + 1) for i in range(10)]
        gate_class_list = Gate.__subclasses__() + ControlledGate.__subclasses__()
        for gate_class in gate_class_list:
            if hasattr(gate_class, "__abstractmethods__"):
                # gate_class is abstract
                continue
            sig = signature(gate_class)
            free_params = len(set(sig.parameters) - {"label", "ctrl_state"})
            try:
                if gate_class == PauliGate:
                    # special case due to PauliGate using string parameters
                    gate = gate_class("IXYZ")
                elif gate_class == BooleanExpression:
                    gate = gate_class("x")
                elif gate_class == PauliEvolutionGate:
                    gate = gate_class(Pauli("XYZ"))
                else:
                    gate = gate_class(*params[0:free_params])
            except (CircuitError, QiskitError, AttributeError, TypeError):
                self.log.info("Cannot init gate with params only. Skipping %s", gate_class)
                continue
            if gate.name in ["U", "CX"]:
                continue
            circ = QuantumCircuit(gate.num_qubits)
            circ.append(gate, range(gate.num_qubits))
            try:
                gate_matrix = gate.to_matrix()
            except CircuitError:
                # gate doesn't implement to_matrix method: skip
                self.log.info('to_matrix method FAILED for "%s" gate', gate.name)
                continue
            definition_unitary = Operator(circ)

            with self.subTest(gate_class):
                # TODO check for exact equality
                self.assertTrue(matrix_equal(definition_unitary, gate_matrix, ignore_phase=True))
                self.assertTrue(is_unitary_matrix(gate_matrix))

    @unittest.skipUnless(HAS_TWEEDLEDUM, "tweedledum required for this test")
    def test_to_matrix_op(self):
        """test gates implementing to_matrix generate matrix which matches
        definition using Operator."""
        from qiskit.circuit.library.generalized_gates.gms import MSGate
        from qiskit.circuit.library.generalized_gates.pauli import PauliGate
        from qiskit.circuit.library.pauli_evolution import PauliEvolutionGate
        from qiskit.circuit.classicalfunction.boolean_expression import BooleanExpression

        params = [0.1 * i for i in range(1, 11)]
        gate_class_list = Gate.__subclasses__() + ControlledGate.__subclasses__()
        for gate_class in gate_class_list:
            if hasattr(gate_class, "__abstractmethods__"):
                # gate_class is abstract
                continue
            sig = signature(gate_class)
            if gate_class == MSGate:
                # due to the signature (num_qubits, theta, *, n_qubits=Noe) the signature detects
                # 3 arguments but really its only 2. This if can be removed once the deprecated
                # n_qubits argument is no longer supported.
                free_params = 2
            else:
                free_params = len(set(sig.parameters) - {"label", "ctrl_state"})
            try:
                if gate_class == PauliGate:
                    # special case due to PauliGate using string parameters
                    gate = gate_class("IXYZ")
                elif gate_class == BooleanExpression:
                    gate = gate_class("x")
                elif gate_class == PauliEvolutionGate:
                    gate = gate_class(Pauli("XYZ"))
                else:
                    gate = gate_class(*params[0:free_params])
            except (CircuitError, QiskitError, AttributeError, TypeError):
                self.log.info("Cannot init gate with params only. Skipping %s", gate_class)
                continue
            if gate.name in ["U", "CX"]:
                continue
            try:
                gate_matrix = gate.to_matrix()
            except CircuitError:
                # gate doesn't implement to_matrix method: skip
                self.log.info('to_matrix method FAILED for "%s" gate', gate.name)
                continue
            if not hasattr(gate, "definition") or not gate.definition:
                continue
            definition_unitary = Operator(gate.definition).data
            self.assertTrue(matrix_equal(definition_unitary, gate_matrix))
            self.assertTrue(is_unitary_matrix(gate_matrix))


if __name__ == "__main__":
    unittest.main(verbosity=2)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""User-space constructor functions for the expression tree, which do some of the inference and
lifting boilerplate work."""

# pylint: disable=redefined-builtin,redefined-outer-name

from __future__ import annotations

__all__ = [
    "lift",
    "bit_not",
    "logic_not",
    "bit_and",
    "bit_or",
    "bit_xor",
    "logic_and",
    "logic_or",
    "equal",
    "not_equal",
    "less",
    "less_equal",
    "greater",
    "greater_equal",
    "lift_legacy_condition",
]

import typing

from .expr import Expr, Var, Value, Unary, Binary, Cast, Index
from ..types import CastKind, cast_kind
from .. import types

if typing.TYPE_CHECKING:
    import qiskit


def _coerce_lossless(expr: Expr, type: types.Type) -> Expr:
    """Coerce ``expr`` to ``type`` by inserting a suitable :class:`Cast` node, if the cast is
    lossless.  Otherwise, raise a ``TypeError``."""
    kind = cast_kind(expr.type, type)
    if kind is CastKind.EQUAL:
        return expr
    if kind is CastKind.IMPLICIT:
        return Cast(expr, type, implicit=True)
    if kind is CastKind.LOSSLESS:
        return Cast(expr, type, implicit=False)
    if kind is CastKind.DANGEROUS:
        raise TypeError(f"cannot cast '{expr}' to '{type}' without loss of precision")
    raise TypeError(f"no cast is defined to take '{expr}' to '{type}'")


def lift_legacy_condition(
    condition: tuple[qiskit.circuit.Clbit | qiskit.circuit.ClassicalRegister, int], /
) -> Expr:
    """Lift a legacy two-tuple equality condition into a new-style :class:`Expr`.

    Examples:
        Taking an old-style conditional instruction and getting an :class:`Expr` from its
        condition::

            from qiskit.circuit import ClassicalRegister
            from qiskit.circuit.library import HGate
            from qiskit.circuit.classical import expr

            cr = ClassicalRegister(2)
            instr = HGate().c_if(cr, 3)

            lifted = expr.lift_legacy_condition(instr.condition)
    """
    from qiskit.circuit import Clbit  # pylint: disable=cyclic-import

    target, value = condition
    if isinstance(target, Clbit):
        bool_ = types.Bool()
        return Var(target, bool_) if value else Unary(Unary.Op.LOGIC_NOT, Var(target, bool_), bool_)
    left = Var(target, types.Uint(width=target.size))
    if value.bit_length() > target.size:
        left = Cast(left, types.Uint(width=value.bit_length()), implicit=True)
    right = Value(value, left.type)
    return Binary(Binary.Op.EQUAL, left, right, types.Bool())


def lift(value: typing.Any, /, type: types.Type | None = None) -> Expr:
    """Lift the given Python ``value`` to a :class:`~.expr.Value` or :class:`~.expr.Var`.

    If an explicit ``type`` is given, the typing in the output will reflect that.

    Examples:
        Lifting simple circuit objects to be :class:`~.expr.Var` instances::

            >>> from qiskit.circuit import Clbit, ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.lift(Clbit())
            Var(<clbit>, Bool())
            >>> expr.lift(ClassicalRegister(3, "c"))
            Var(ClassicalRegister(3, "c"), Uint(3))

        The type of the return value can be influenced, if the given value could be interpreted
        losslessly as the given type (use :func:`cast` to perform a full set of casting
        operations, include lossy ones)::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr, types
            >>> expr.lift(ClassicalRegister(3, "c"), types.Uint(5))
            Var(ClassicalRegister(3, "c"), Uint(5))
            >>> expr.lift(5, types.Uint(4))
            Value(5, Uint(4))
    """
    if isinstance(value, Expr):
        if type is not None:
            raise ValueError("use 'cast' to cast existing expressions, not 'lift'")
        return value
    from qiskit.circuit import Clbit, ClassicalRegister  # pylint: disable=cyclic-import

    inferred: types.Type
    if value is True or value is False or isinstance(value, Clbit):
        inferred = types.Bool()
        constructor = Value if value is True or value is False else Var
    elif isinstance(value, ClassicalRegister):
        inferred = types.Uint(width=value.size)
        constructor = Var
    elif isinstance(value, int):
        if value < 0:
            raise ValueError("cannot represent a negative value")
        inferred = types.Uint(width=value.bit_length() or 1)
        constructor = Value
    else:
        raise TypeError(f"failed to infer a type for '{value}'")
    if type is None:
        type = inferred
    if types.is_supertype(type, inferred):
        return constructor(value, type)
    raise TypeError(
        f"the explicit type '{type}' is not suitable for representing '{value}';"
        f" it must be non-strict supertype of '{inferred}'"
    )


def cast(operand: typing.Any, type: types.Type, /) -> Expr:
    """Create an explicit cast from the given value to the given type.

    Examples:
        Add an explicit cast node that explicitly casts a higher precision type to a lower precision
        one::

            >>> from qiskit.circuit.classical import expr, types
            >>> value = expr.value(5, types.Uint(32))
            >>> expr.cast(value, types.Uint(8))
            Cast(Value(5, types.Uint(32)), types.Uint(8), implicit=False)
    """
    operand = lift(operand)
    if cast_kind(operand.type, type) is CastKind.NONE:
        raise TypeError(f"cannot cast '{operand}' to '{type}'")
    return Cast(operand, type)


def bit_not(operand: typing.Any, /) -> Expr:
    """Create a bitwise 'not' expression node from the given value, resolving any implicit casts and
    lifting the value into a :class:`Value` node if required.

    Examples:
        Bitwise negation of a :class:`.ClassicalRegister`::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.bit_not(ClassicalRegister(3, "c"))
            Unary(Unary.Op.BIT_NOT, Var(ClassicalRegister(3, 'c'), Uint(3)), Uint(3))
    """
    operand = lift(operand)
    if operand.type.kind not in (types.Bool, types.Uint):
        raise TypeError(f"cannot apply '{Unary.Op.BIT_NOT}' to type '{operand.type}'")
    return Unary(Unary.Op.BIT_NOT, operand, operand.type)


def logic_not(operand: typing.Any, /) -> Expr:
    """Create a logical 'not' expression node from the given value, resolving any implicit casts and
    lifting the value into a :class:`Value` node if required.

    Examples:
        Logical negation of a :class:`.ClassicalRegister`::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.logic_not(ClassicalRegister(3, "c"))
            Unary(\
Unary.Op.LOGIC_NOT, \
Cast(Var(ClassicalRegister(3, 'c'), Uint(3)), Bool(), implicit=True), \
Bool())
    """
    operand = _coerce_lossless(lift(operand), types.Bool())
    return Unary(Unary.Op.LOGIC_NOT, operand, operand.type)


def _lift_binary_operands(left: typing.Any, right: typing.Any) -> tuple[Expr, Expr]:
    """Lift two binary operands simultaneously, inferring the widths of integer literals in either
    position to match the other operand."""
    left_int = isinstance(left, int) and not isinstance(left, bool)
    right_int = isinstance(right, int) and not isinstance(right, bool)
    if not (left_int or right_int):
        left = lift(left)
        right = lift(right)
    elif not right_int:
        right = lift(right)
        if right.type.kind is types.Uint:
            if left.bit_length() > right.type.width:
                raise TypeError(
                    f"integer literal '{left}' is wider than the other operand '{right}'"
                )
            left = Value(left, right.type)
        else:
            left = lift(left)
    elif not left_int:
        left = lift(left)
        if left.type.kind is types.Uint:
            if right.bit_length() > left.type.width:
                raise TypeError(
                    f"integer literal '{right}' is wider than the other operand '{left}'"
                )
            right = Value(right, left.type)
        else:
            right = lift(right)
    else:
        # Both are `int`, so we take our best case to make things work.
        uint = types.Uint(max(left.bit_length(), right.bit_length(), 1))
        left = Value(left, uint)
        right = Value(right, uint)
    return left, right


def _binary_bitwise(op: Binary.Op, left: typing.Any, right: typing.Any) -> Expr:
    left, right = _lift_binary_operands(left, right)
    type: types.Type
    if left.type.kind is right.type.kind is types.Bool:
        type = types.Bool()
    elif left.type.kind is types.Uint and right.type.kind is types.Uint:
        if left.type != right.type:
            raise TypeError(
                "binary bitwise operations are defined between unsigned integers of the same width,"
                f" but got {left.type.width} and {right.type.width}."
            )
        type = left.type
    else:
        raise TypeError(f"invalid types for '{op}': '{left.type}' and '{right.type}'")
    return Binary(op, left, right, type)


def bit_and(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create a bitwise 'and' expression node from the given value, resolving any implicit casts and
    lifting the values into :class:`Value` nodes if required.

    Examples:
        Bitwise 'and' of a classical register and an integer literal::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.bit_and(ClassicalRegister(3, "c"), 0b111)
            Binary(\
Binary.Op.BIT_AND, \
Var(ClassicalRegister(3, 'c'), Uint(3)), \
Value(7, Uint(3)), \
Uint(3))
        """
    return _binary_bitwise(Binary.Op.BIT_AND, left, right)


def bit_or(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create a bitwise 'or' expression node from the given value, resolving any implicit casts and
    lifting the values into :class:`Value` nodes if required.

    Examples:
        Bitwise 'or' of a classical register and an integer literal::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.bit_or(ClassicalRegister(3, "c"), 0b101)
            Binary(\
Binary.Op.BIT_OR, \
Var(ClassicalRegister(3, 'c'), Uint(3)), \
Value(5, Uint(3)), \
Uint(3))
    """
    return _binary_bitwise(Binary.Op.BIT_OR, left, right)


def bit_xor(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create a bitwise 'exclusive or' expression node from the given value, resolving any implicit
    casts and lifting the values into :class:`Value` nodes if required.

    Examples:
        Bitwise 'exclusive or' of a classical register and an integer literal::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.bit_xor(ClassicalRegister(3, "c"), 0b101)
            Binary(\
Binary.Op.BIT_XOR, \
Var(ClassicalRegister(3, 'c'), Uint(3)), \
Value(5, Uint(3)), \
Uint(3))
    """
    return _binary_bitwise(Binary.Op.BIT_XOR, left, right)


def _binary_logical(op: Binary.Op, left: typing.Any, right: typing.Any) -> Expr:
    bool_ = types.Bool()
    left = _coerce_lossless(lift(left), bool_)
    right = _coerce_lossless(lift(right), bool_)
    return Binary(op, left, right, bool_)


def logic_and(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create a logical 'and' expression node from the given value, resolving any implicit casts and
    lifting the values into :class:`Value` nodes if required.

    Examples:
        Logical 'and' of two classical bits::

            >>> from qiskit.circuit import Clbit
            >>> from qiskit.circuit.classical import expr
            >>> expr.logical_and(Clbit(), Clbit())
            Binary(Binary.Op.LOGIC_AND, Var(<clbit 0>, Bool()), Var(<clbit 1>, Bool()), Bool())
    """
    return _binary_logical(Binary.Op.LOGIC_AND, left, right)


def logic_or(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create a logical 'or' expression node from the given value, resolving any implicit casts and
    lifting the values into :class:`Value` nodes if required.

    Examples:
        Logical 'or' of two classical bits

            >>> from qiskit.circuit import Clbit
            >>> from qiskit.circuit.classical import expr
            >>> expr.logical_and(Clbit(), Clbit())
            Binary(Binary.Op.LOGIC_OR, Var(<clbit 0>, Bool()), Var(<clbit 1>, Bool()), Bool())
    """
    return _binary_logical(Binary.Op.LOGIC_OR, left, right)


def _equal_like(op: Binary.Op, left: typing.Any, right: typing.Any) -> Expr:
    left, right = _lift_binary_operands(left, right)
    if left.type.kind is not right.type.kind:
        raise TypeError(f"invalid types for '{op}': '{left.type}' and '{right.type}'")
    type = types.greater(left.type, right.type)
    return Binary(op, _coerce_lossless(left, type), _coerce_lossless(right, type), types.Bool())


def equal(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create an 'equal' expression node from the given value, resolving any implicit casts and
    lifting the values into :class:`Value` nodes if required.

    Examples:
        Equality between a classical register and an integer::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.equal(ClassicalRegister(3, "c"), 7)
            Binary(Binary.Op.EQUAL, \
Var(ClassicalRegister(3, "c"), Uint(3)), \
Value(7, Uint(3)), \
Uint(3))
    """
    return _equal_like(Binary.Op.EQUAL, left, right)


def not_equal(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create a 'not equal' expression node from the given value, resolving any implicit casts and
    lifting the values into :class:`Value` nodes if required.

    Examples:
        Inequality between a classical register and an integer::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.not_equal(ClassicalRegister(3, "c"), 7)
            Binary(Binary.Op.NOT_EQUAL, \
Var(ClassicalRegister(3, "c"), Uint(3)), \
Value(7, Uint(3)), \
Uint(3))
    """
    return _equal_like(Binary.Op.NOT_EQUAL, left, right)


def _binary_relation(op: Binary.Op, left: typing.Any, right: typing.Any) -> Expr:
    left, right = _lift_binary_operands(left, right)
    if left.type.kind is not right.type.kind or left.type.kind is types.Bool:
        raise TypeError(f"invalid types for '{op}': '{left.type}' and '{right.type}'")
    type = types.greater(left.type, right.type)
    return Binary(op, _coerce_lossless(left, type), _coerce_lossless(right, type), types.Bool())


def less(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create a 'less than' expression node from the given value, resolving any implicit casts and
    lifting the values into :class:`Value` nodes if required.

    Examples:
        Query if a classical register is less than an integer::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.less(ClassicalRegister(3, "c"), 5)
            Binary(Binary.Op.LESS, \
Var(ClassicalRegister(3, "c"), Uint(3)), \
Value(5, Uint(3)), \
Uint(3))
    """
    return _binary_relation(Binary.Op.LESS, left, right)


def less_equal(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create a 'less than or equal to' expression node from the given value, resolving any implicit
    casts and lifting the values into :class:`Value` nodes if required.

    Examples:
        Query if a classical register is less than or equal to another::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.less(ClassicalRegister(3, "a"), ClassicalRegister(3, "b"))
            Binary(Binary.Op.LESS_EQUAL, \
Var(ClassicalRegister(3, "a"), Uint(3)), \
Var(ClassicalRegister(3, "b"), Uint(3)), \
Uint(3))
    """
    return _binary_relation(Binary.Op.LESS_EQUAL, left, right)


def greater(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create a 'greater than' expression node from the given value, resolving any implicit casts
    and lifting the values into :class:`Value` nodes if required.

    Examples:
        Query if a classical register is greater than an integer::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.less(ClassicalRegister(3, "c"), 5)
            Binary(Binary.Op.GREATER, \
Var(ClassicalRegister(3, "c"), Uint(3)), \
Value(5, Uint(3)), \
Uint(3))
    """
    return _binary_relation(Binary.Op.GREATER, left, right)


def greater_equal(left: typing.Any, right: typing.Any, /) -> Expr:
    """Create a 'greater than or equal to' expression node from the given value, resolving any
    implicit casts and lifting the values into :class:`Value` nodes if required.

    Examples:
        Query if a classical register is greater than or equal to another::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.less(ClassicalRegister(3, "a"), ClassicalRegister(3, "b"))
            Binary(Binary.Op.GREATER_EQUAL, \
Var(ClassicalRegister(3, "a"), Uint(3)), \
Var(ClassicalRegister(3, "b"), Uint(3)), \
Uint(3))
    """
    return _binary_relation(Binary.Op.GREATER_EQUAL, left, right)


def _shift_like(
    op: Binary.Op, left: typing.Any, right: typing.Any, type: types.Type | None
) -> Expr:
    if type is not None and type.kind is not types.Uint:
        raise TypeError(f"type '{type}' is not a valid bitshift operand type")
    if isinstance(left, Expr):
        left = _coerce_lossless(left, type) if type is not None else left
    else:
        left = lift(left, type)
    right = lift(right)
    if left.type.kind != types.Uint or right.type.kind != types.Uint:
        raise TypeError(f"invalid types for '{op}': '{left.type}' and '{right.type}'")
    return Binary(op, left, right, left.type)


def shift_left(left: typing.Any, right: typing.Any, /, type: types.Type | None = None) -> Expr:
    """Create a 'bitshift left' expression node from the given two values, resolving any implicit
    casts and lifting the values into :class:`Value` nodes if required.

    If ``type`` is given, the ``left`` operand will be coerced to it (if possible).

    Examples:
        Shift the value of a standalone variable left by some amount::

            >>> from qiskit.circuit.classical import expr, types
            >>> a = expr.Var.new("a", types.Uint(8))
            >>> expr.shift_left(a, 4)
            Binary(Binary.Op.SHIFT_LEFT, \
Var(<UUID>, Uint(8), name='a'), \
Value(4, Uint(3)), \
Uint(8))

        Shift an integer literal by a variable amount, coercing the type of the literal::

            >>> expr.shift_left(3, a, types.Uint(16))
            Binary(Binary.Op.SHIFT_LEFT, \
Value(3, Uint(16)), \
Var(<UUID>, Uint(8), name='a'), \
Uint(16))
    """
    return _shift_like(Binary.Op.SHIFT_LEFT, left, right, type)


def shift_right(left: typing.Any, right: typing.Any, /, type: types.Type | None = None) -> Expr:
    """Create a 'bitshift right' expression node from the given values, resolving any implicit casts
    and lifting the values into :class:`Value` nodes if required.

    If ``type`` is given, the ``left`` operand will be coerced to it (if possible).

    Examples:
        Shift the value of a classical register right by some amount::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.shift_right(ClassicalRegister(8, "a"), 4)
            Binary(Binary.Op.SHIFT_RIGHT, \
Var(ClassicalRegister(8, "a"), Uint(8)), \
Value(4, Uint(3)), \
Uint(8))
    """
    return _shift_like(Binary.Op.SHIFT_RIGHT, left, right, type)


def index(target: typing.Any, index: typing.Any, /) -> Expr:
    """Index into the ``target`` with the given integer ``index``, lifting the values into
    :class:`Value` nodes if required.

    This can be used as the target of a :class:`.Store`, if the ``target`` is itself an lvalue.

    Examples:
        Index into a classical register with a literal::

            >>> from qiskit.circuit import ClassicalRegister
            >>> from qiskit.circuit.classical import expr
            >>> expr.index(ClassicalRegister(8, "a"), 3)
            Index(Var(ClassicalRegister(8, "a"), Uint(8)), Value(3, Uint(2)), Bool())
    """
    target, index = lift(target), lift(index)
    if target.type.kind is not types.Uint or index.type.kind is not types.Uint:
        raise TypeError(f"invalid types for indexing: '{target.type}' and '{index.type}'")
    return Index(target, index, types.Bool())



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test library of phase estimation circuits."""

import unittest
import numpy as np

from qiskit.circuit import QuantumCircuit
from qiskit.circuit.library import PhaseEstimation, QFT, phase_estimation
from qiskit.quantum_info import Statevector
from qiskit.compiler import transpile
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestPhaseEstimation(QiskitTestCase):
    """Test the phase estimation circuit."""

    def assertPhaseEstimationIsCorrect(
        self, pec: QuantumCircuit, eigenstate: QuantumCircuit, phase_as_binary: str
    ):
        r"""Assert that the phase estimation circuit implements the correct transformation.

        Applying the phase estimation circuit on a target register which holds the eigenstate
        :math:`|u\rangle` (say the last register), the final state should be

        .. math::

            |\phi_1\rangle \cdots |\phi_t\rangle |u\rangle

        where the eigenvalue is written as :math:`e^{2\pi i \phi}` and the angle is represented
        in binary fraction, i.e. :math:`\phi = 0.\phi_1 \ldots \phi_t`.

        Args:
            pec: The circuit implementing the phase estimation circuit.
            eigenstate: The eigenstate as circuit.
            phase_as_binary: The phase of the eigenvalue in a binary fraction. E.g. if the
                phase is 0.25, the binary fraction is '01' as 0.01 = 0 * 0.5 + 1 * 0.25 = 0.25.
        """

        # the target state
        eigenstate_as_vector = Statevector.from_instruction(eigenstate).data
        reference = eigenstate_as_vector

        zero, one = [1, 0], [0, 1]
        for qubit in phase_as_binary[::-1]:
            reference = np.kron(reference, zero if qubit == "0" else one)

        # the simulated state
        circuit = QuantumCircuit(pec.num_qubits)
        circuit.compose(
            eigenstate,
            list(range(pec.num_qubits - eigenstate.num_qubits, pec.num_qubits)),
            inplace=True,
        )
        circuit.compose(pec, inplace=True)
        actual = Statevector(circuit)
        np.testing.assert_almost_equal(reference, actual)

    def test_phase_estimation(self):
        """Test the standard phase estimation circuit."""
        with self.subTest("U=S, psi=|1>"):
            unitary = QuantumCircuit(1)
            unitary.s(0)

            eigenstate = QuantumCircuit(1)
            eigenstate.x(0)

            # eigenvalue is 1j = exp(2j pi 0.25) thus phi = 0.25 = 0.010 = '010'
            # using three digits as 3 evaluation qubits are used
            phase_as_binary = "0100"

            pec = PhaseEstimation(4, unitary)

            self.assertPhaseEstimationIsCorrect(pec, eigenstate, phase_as_binary)

        with self.subTest("U=SZ, psi=|11>"):
            unitary = QuantumCircuit(2)
            unitary.z(0)
            unitary.s(1)

            eigenstate = QuantumCircuit(2)
            eigenstate.x([0, 1])

            # eigenvalue is -1j = exp(2j pi 0.75) thus phi = 0.75 = 0.110 = '110'
            # using three digits as 3 evaluation qubits are used
            phase_as_binary = "110"

            pec = PhaseEstimation(3, unitary)

            self.assertPhaseEstimationIsCorrect(pec, eigenstate, phase_as_binary)

        with self.subTest("a 3-q unitary"):
            #      ┌───┐
            # q_0: ┤ X ├──■────■───────
            #      ├───┤  │    │
            # q_1: ┤ X ├──■────■───────
            #      ├───┤┌───┐┌─┴─┐┌───┐
            # q_2: ┤ X ├┤ H ├┤ X ├┤ H ├
            #      └───┘└───┘└───┘└───┘
            unitary = QuantumCircuit(3)
            unitary.x([0, 1, 2])
            unitary.cz(0, 1)
            unitary.h(2)
            unitary.ccx(0, 1, 2)
            unitary.h(2)

            #      ┌───┐
            # q_0: ┤ H ├──■────■──
            #      └───┘┌─┴─┐  │
            # q_1: ─────┤ X ├──┼──
            #           └───┘┌─┴─┐
            # q_2: ──────────┤ X ├
            #                └───┘
            eigenstate = QuantumCircuit(3)
            eigenstate.h(0)
            eigenstate.cx(0, 1)
            eigenstate.cx(0, 2)

            # the unitary acts as identity on the eigenstate, thus the phase is 0
            phase_as_binary = "00"

            pec = PhaseEstimation(2, unitary)

            self.assertPhaseEstimationIsCorrect(pec, eigenstate, phase_as_binary)

    def test_phase_estimation_iqft_setting(self):
        """Test default and custom setting of the QFT circuit."""
        unitary = QuantumCircuit(1)
        unitary.s(0)

        with self.subTest("default QFT"):
            pec = PhaseEstimation(3, unitary)
            expected_qft = QFT(3, inverse=True, do_swaps=False)
            self.assertEqual(
                pec.decompose().data[-1].operation.definition, expected_qft.decompose()
            )

        with self.subTest("custom QFT"):
            iqft = QFT(3, approximation_degree=2).inverse()
            pec = PhaseEstimation(3, unitary, iqft=iqft)
            self.assertEqual(pec.decompose().data[-1].operation.definition, iqft.decompose())

    def test_phase_estimation_function(self):
        """Test the phase estimation function."""
        with self.subTest("U=S, psi=|1>"):
            unitary = QuantumCircuit(1)
            unitary.s(0)

            eigenstate = QuantumCircuit(1)
            eigenstate.x(0)

            # eigenvalue is 1j = exp(2j pi 0.25) thus phi = 0.25 = 0.010 = '010'
            # using three digits as 3 evaluation qubits are used
            phase_as_binary = "0100"

            pec = phase_estimation(4, unitary)

            self.assertPhaseEstimationIsCorrect(pec, eigenstate, phase_as_binary)

        with self.subTest("U=SZ, psi=|11>"):
            unitary = QuantumCircuit(2)
            unitary.z(0)
            unitary.s(1)

            eigenstate = QuantumCircuit(2)
            eigenstate.x([0, 1])

            # eigenvalue is -1j = exp(2j pi 0.75) thus phi = 0.75 = 0.110 = '110'
            # using three digits as 3 evaluation qubits are used
            phase_as_binary = "110"

            pec = phase_estimation(3, unitary)

            self.assertPhaseEstimationIsCorrect(pec, eigenstate, phase_as_binary)

        with self.subTest("a 3-q unitary"):
            #      ┌───┐
            # q_0: ┤ X ├──■────■───────
            #      ├───┤  │    │
            # q_1: ┤ X ├──■────■───────
            #      ├───┤┌───┐┌─┴─┐┌───┐
            # q_2: ┤ X ├┤ H ├┤ X ├┤ H ├
            #      └───┘└───┘└───┘└───┘
            unitary = QuantumCircuit(3)
            unitary.x([0, 1, 2])
            unitary.cz(0, 1)
            unitary.h(2)
            unitary.ccx(0, 1, 2)
            unitary.h(2)

            #      ┌───┐
            # q_0: ┤ H ├──■────■──
            #      └───┘┌─┴─┐  │
            # q_1: ─────┤ X ├──┼──
            #           └───┘┌─┴─┐
            # q_2: ──────────┤ X ├
            #                └───┘
            eigenstate = QuantumCircuit(3)
            eigenstate.h(0)
            eigenstate.cx(0, 1)
            eigenstate.cx(0, 2)

            # the unitary acts as identity on the eigenstate, thus the phase is 0
            phase_as_binary = "00"

            pec = phase_estimation(2, unitary)

            self.assertPhaseEstimationIsCorrect(pec, eigenstate, phase_as_binary)

    def test_phase_estimation_function_swaps_get_removed(self):
        """Test that transpiling the circuit correctly removes swaps and permutations."""
        unitary = QuantumCircuit(1)
        unitary.s(0)

        eigenstate = QuantumCircuit(1)
        eigenstate.x(0)

        pec = phase_estimation(4, unitary)

        # transpilation (or more precisely HighLevelSynthesis + ElidePermutations) should
        # remove all swap gates (possibly added when synthesizing the QFTGate in the circuit)
        # and the final permutation gate
        transpiled = transpile(pec)
        transpiled_ops = transpiled.count_ops()

        self.assertNotIn("permutation", transpiled_ops)
        self.assertNotIn("swap", transpiled_ops)
        self.assertNotIn("cx", transpiled_ops)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Tests AQC framework using hardcoded and randomly generated circuits.
"""
from functools import partial
import unittest
from ddt import ddt, data
import numpy as np
from scipy.optimize import minimize

from qiskit.quantum_info import Operator
from qiskit.synthesis.unitary.aqc.aqc import AQC
from qiskit.synthesis.unitary.aqc.cnot_structures import make_cnot_network
from qiskit.synthesis.unitary.aqc.cnot_unit_circuit import CNOTUnitCircuit
from qiskit.synthesis.unitary.aqc.cnot_unit_objective import DefaultCNOTUnitObjective
from qiskit.synthesis.unitary.aqc.fast_gradient.fast_gradient import FastCNOTUnitObjective
from test import QiskitTestCase  # pylint: disable=wrong-import-order
from test.python.synthesis.aqc.sample_data import (  # pylint: disable=wrong-import-order
    ORIGINAL_CIRCUIT,
    INITIAL_THETAS,
)


@ddt
class TestAqc(QiskitTestCase):
    """Main tests of approximate quantum compiler."""

    @data(True, False)
    def test_aqc(self, uses_default):
        """Tests AQC on a hardcoded circuit/matrix."""

        seed = 12345

        num_qubits = int(round(np.log2(ORIGINAL_CIRCUIT.shape[0])))
        cnots = make_cnot_network(
            num_qubits=num_qubits, network_layout="spin", connectivity_type="full", depth=0
        )

        if uses_default:
            aqc = AQC(seed=seed)
        else:
            optimizer = partial(minimize, args=(), method="L-BFGS-B", options={"maxiter": 200})
            aqc = AQC(optimizer=optimizer, seed=seed)

        target_matrix = ORIGINAL_CIRCUIT
        approximate_circuit = CNOTUnitCircuit(num_qubits, cnots)
        approximating_objective = DefaultCNOTUnitObjective(num_qubits, cnots)

        aqc.compile_unitary(
            target_matrix=target_matrix,
            approximate_circuit=approximate_circuit,
            approximating_objective=approximating_objective,
            initial_point=INITIAL_THETAS,
        )

        approx_matrix = Operator(approximate_circuit).data
        error = 0.5 * (np.linalg.norm(approx_matrix - ORIGINAL_CIRCUIT, "fro") ** 2)
        self.assertLess(error, 1e-3)

    def test_aqc_fastgrad(self):
        """
        Tests AQC on a MCX circuit/matrix with random initial guess using
        the accelerated implementation.
        """
        seed = 12345
        np.random.seed(seed)

        num_qubits = int(3)
        cnots = make_cnot_network(
            num_qubits=num_qubits, network_layout="spin", connectivity_type="full", depth=0
        )

        optimizer = partial(minimize, args=(), method="L-BFGS-B", options={"maxiter": 200})
        aqc = AQC(optimizer=optimizer, seed=seed)

        # Make multi-control CNOT gate matrix.
        # Another option: target_matrix = ORIGINAL_CIRCUIT
        target_matrix = np.eye(2**num_qubits, dtype=np.complex128)
        target_matrix[-2:, -2:] = [[0, 1], [1, 0]]

        circ = CNOTUnitCircuit(num_qubits, cnots)
        objv = FastCNOTUnitObjective(num_qubits, cnots)

        aqc.compile_unitary(
            target_matrix=target_matrix,
            approximate_circuit=circ,
            approximating_objective=objv,
            initial_point=2 * np.pi * np.random.rand(objv.num_thetas),
        )

        approx_matrix = Operator(circ).data
        error = 0.5 * (np.linalg.norm(approx_matrix - target_matrix, "fro") ** 2)
        self.assertTrue(error < 1e-3)

    def test_aqc_determinant_minus_one(self):
        """
        Tests AQC on a matrix with determinant −1.
        """
        seed = 12345

        num_qubits = int(3)
        cnots = make_cnot_network(
            num_qubits=num_qubits, network_layout="spin", connectivity_type="full", depth=0
        )

        optimizer = partial(minimize, args=(), method="L-BFGS-B", options={"maxiter": 200})
        aqc = AQC(optimizer=optimizer, seed=seed)

        target_matrix = np.eye(2**num_qubits, dtype=int)
        # Make a unitary with determinant −1 by swapping any two columns
        target_matrix[:, 2], target_matrix[:, 3] = target_matrix[:, 3], target_matrix[:, 2].copy()

        approximate_circuit = CNOTUnitCircuit(num_qubits, cnots)
        approximating_objective = DefaultCNOTUnitObjective(num_qubits, cnots)

        aqc.compile_unitary(
            target_matrix=target_matrix,
            approximate_circuit=approximate_circuit,
            approximating_objective=approximating_objective,
            initial_point=INITIAL_THETAS,
        )

        approx_matrix = Operator(approximate_circuit).data

        error = 0.5 * (np.linalg.norm(approx_matrix - target_matrix, "fro") ** 2)
        self.assertTrue(error < 1e-3)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Diagonal gate tests."""

import unittest
import numpy as np

from qiskit import QuantumCircuit, assemble
from qiskit import QiskitError
from qiskit.compiler import transpile
from qiskit.circuit.library.generalized_gates import DiagonalGate
from qiskit.quantum_info.operators.predicates import matrix_equal
from qiskit.quantum_info import Operator
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestDiagonalGate(QiskitTestCase):
    """
    Diagonal gate tests.
    """

    def test_diag_gate(self):
        """Test diagonal gates."""
        for phases in [
            [0, 0],
            [0, 0.8],
            [0, 0, 1, 1],
            [0, 1, 0.5, 1],
            (2 * np.pi * np.random.rand(2**3)).tolist(),
            (2 * np.pi * np.random.rand(2**4)).tolist(),
            (2 * np.pi * np.random.rand(2**5)).tolist(),
        ]:
            with self.subTest(phases=phases):
                diag = [np.exp(1j * ph) for ph in phases]
                num_qubits = int(np.log2(len(diag)))
                qc = QuantumCircuit(num_qubits)
                gate = DiagonalGate(diag)
                qc.append(gate, qc.qubits)

                # Decompose the gate
                qc = transpile(qc, basis_gates=["u1", "u3", "u2", "cx", "id"], optimization_level=0)
                # Simulate the decomposed gate
                unitary = Operator(qc)
                unitary_desired = _get_diag_gate_matrix(diag)
                self.assertTrue(matrix_equal(unitary, unitary_desired, ignore_phase=False))

    def test_mod1_entries(self):
        """Test that diagonal raises if entries do not have modules of 1."""
        from qiskit.quantum_info.operators.predicates import ATOL_DEFAULT, RTOL_DEFAULT

        with self.assertRaises(QiskitError):
            DiagonalGate([1, 1 - 2 * ATOL_DEFAULT - RTOL_DEFAULT])

    def test_npcomplex_params_conversion(self):
        """Verify diagonal gate converts numpy.complex to complex."""
        # ref: https://github.com/Qiskit/qiskit-aer/issues/696
        diag = np.array([1 + 0j, 1 + 0j])
        qc = QuantumCircuit(1)
        gate = DiagonalGate(diag)
        qc.append(gate, [0])

        params = qc.data[0].operation.params
        self.assertTrue(
            all(isinstance(p, complex) and not isinstance(p, np.number) for p in params)
        )

        with self.assertWarns(DeprecationWarning):
            # REMOVE this assertion (not the full test) once ASSEMBLE is removed.
            qobj = assemble(qc)
            params = qobj.experiments[0].instructions[0].params
            self.assertTrue(
                all(isinstance(p, complex) and not isinstance(p, np.number) for p in params)
            )

    def test_repeat(self):
        """Test the repeat() method."""
        for phases in [
            [0, 0],
            np.array([0, 0.8, 1, 0]),
            (2 * np.pi * np.random.rand(2**3)).tolist(),
        ]:
            with self.subTest(phases=phases):
                diag = [np.exp(1j * ph) for ph in phases]
                gate = DiagonalGate(diag)
                operator = Operator(gate)
                self.assertTrue(np.allclose(Operator(gate.repeat(2)), operator @ operator))


def _get_diag_gate_matrix(diag):
    return np.diagflat(diag)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-docstring

import os
import pathlib
import re
import shutil
import tempfile
import unittest
import warnings
from unittest.mock import patch

from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister, visualization
from qiskit.utils import optionals
from qiskit.visualization.circuit import styles, text
from qiskit.visualization.exceptions import VisualizationError
from test import QiskitTestCase  # pylint: disable=wrong-import-order

if optionals.HAS_MATPLOTLIB:
    from matplotlib import figure
if optionals.HAS_PIL:
    from PIL import Image


_latex_drawer_condition = unittest.skipUnless(
    all(
        (
            optionals.HAS_PYLATEX,
            optionals.HAS_PIL,
            optionals.HAS_PDFLATEX,
            optionals.HAS_PDFTOCAIRO,
        )
    ),
    "Skipped because not all of PIL, pylatex, pdflatex and pdftocairo are available",
)


class TestCircuitDrawer(QiskitTestCase):
    def test_default_output(self):
        with patch("qiskit.user_config.get_config", return_value={}):
            circuit = QuantumCircuit()
            out = visualization.circuit_drawer(circuit)
            self.assertIsInstance(out, text.TextDrawing)

    @unittest.skipUnless(optionals.HAS_MATPLOTLIB, "Skipped because matplotlib is not available")
    def test_mpl_config_with_path(self):
        # pylint: disable=possibly-used-before-assignment
        # It's too easy to get too nested in a test with many context managers.
        tempdir = tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
        self.addCleanup(tempdir.cleanup)

        clifford_style = pathlib.Path(styles.__file__).parent / "clifford.json"
        shutil.copyfile(clifford_style, pathlib.Path(tempdir.name) / "my_clifford.json")

        circuit = QuantumCircuit(1)
        circuit.h(0)

        def config(style_name):
            return {
                "circuit_drawer": "mpl",
                "circuit_mpl_style": style_name,
                "circuit_mpl_style_path": [tempdir.name],
            }

        with warnings.catch_warnings():
            warnings.filterwarnings("error", message="Style JSON file.*not found")

            # Test that a non-standard style can be loaded by name.
            with patch("qiskit.user_config.get_config", return_value=config("my_clifford")):
                self.assertIsInstance(visualization.circuit_drawer(circuit), figure.Figure)

            # Test that a non-existent style issues a warning, but still draws something.
            with patch("qiskit.user_config.get_config", return_value=config("NONEXISTENT")):
                with self.assertWarnsRegex(UserWarning, "Style JSON file.*not found"):
                    fig = visualization.circuit_drawer(circuit)
                self.assertIsInstance(fig, figure.Figure)

    @unittest.skipUnless(optionals.HAS_MATPLOTLIB, "Skipped because matplotlib is not available")
    def test_user_config_default_output(self):
        with patch("qiskit.user_config.get_config", return_value={"circuit_drawer": "mpl"}):
            circuit = QuantumCircuit()
            out = visualization.circuit_drawer(circuit)
            self.assertIsInstance(out, figure.Figure)

    def test_default_output_with_user_config_not_set(self):
        with patch("qiskit.user_config.get_config", return_value={"other_option": True}):
            circuit = QuantumCircuit()
            out = visualization.circuit_drawer(circuit)
            self.assertIsInstance(out, text.TextDrawing)

    @unittest.skipUnless(optionals.HAS_MATPLOTLIB, "Skipped because matplotlib is not available")
    def test_kwarg_priority_over_user_config_default_output(self):
        with patch("qiskit.user_config.get_config", return_value={"circuit_drawer": "latex"}):
            circuit = QuantumCircuit()
            out = visualization.circuit_drawer(circuit, output="mpl")
            self.assertIsInstance(out, figure.Figure)

    @unittest.skipUnless(optionals.HAS_MATPLOTLIB, "Skipped because matplotlib is not available")
    def test_default_backend_auto_output_with_mpl(self):
        with patch("qiskit.user_config.get_config", return_value={"circuit_drawer": "auto"}):
            circuit = QuantumCircuit()
            out = visualization.circuit_drawer(circuit)
            self.assertIsInstance(out, figure.Figure)

    def test_default_backend_auto_output_without_mpl(self):
        with patch("qiskit.user_config.get_config", return_value={"circuit_drawer": "auto"}):
            with optionals.HAS_MATPLOTLIB.disable_locally():
                circuit = QuantumCircuit()
                out = visualization.circuit_drawer(circuit)
                self.assertIsInstance(out, text.TextDrawing)

    @_latex_drawer_condition
    def test_latex_unsupported_image_format_error_message(self):
        with patch("qiskit.user_config.get_config", return_value={"circuit_drawer": "latex"}):
            circuit = QuantumCircuit()
            with self.assertRaises(VisualizationError, msg="Pillow could not write the image file"):
                visualization.circuit_drawer(circuit, filename="file.spooky")

    @_latex_drawer_condition
    def test_latex_output_file_correct_format(self):
        # pylint: disable=possibly-used-before-assignment
        with patch("qiskit.user_config.get_config", return_value={"circuit_drawer": "latex"}):
            circuit = QuantumCircuit()
            filename = "file.gif"
            visualization.circuit_drawer(circuit, filename=filename)
            with Image.open(filename) as im:
                if filename.endswith("jpg"):
                    self.assertIn(im.format.lower(), "jpeg")
                else:
                    self.assertIn(im.format.lower(), filename.rsplit(".", maxsplit=1)[-1])
            os.remove(filename)

    def test_wire_order(self):
        """Test wire_order
        See: https://github.com/Qiskit/qiskit-terra/pull/9893"""
        qr = QuantumRegister(4, "q")
        cr = ClassicalRegister(4, "c")
        cr2 = ClassicalRegister(2, "ca")
        circuit = QuantumCircuit(qr, cr, cr2)
        circuit.h(0)
        circuit.h(3)
        circuit.x(1)
        with self.assertWarns(DeprecationWarning):
            circuit.x(3).c_if(cr, 10)

        expected = "\n".join(
            [
                "                  ",
                " q_2: ────────────",
                "      ┌───┐ ┌───┐ ",
                " q_3: ┤ H ├─┤ X ├─",
                "      ├───┤ └─╥─┘ ",
                " q_0: ┤ H ├───╫───",
                "      ├───┤   ║   ",
                " q_1: ┤ X ├───╫───",
                "      └───┘┌──╨──┐",
                " c: 4/═════╡ 0xa ╞",
                "           └─────┘",
                "ca: 2/════════════",
                "                  ",
            ]
        )
        result = visualization.circuit_drawer(circuit, output="text", wire_order=[2, 3, 0, 1])
        self.assertEqual(result.__str__(), expected)

    def test_wire_order_cregbundle(self):
        """Test wire_order with cregbundle=True
        See: https://github.com/Qiskit/qiskit-terra/pull/9893"""
        qr = QuantumRegister(4, "q")
        cr = ClassicalRegister(4, "c")
        cr2 = ClassicalRegister(2, "ca")
        circuit = QuantumCircuit(qr, cr, cr2)
        circuit.h(0)
        circuit.h(3)
        circuit.x(1)
        with self.assertWarns(DeprecationWarning):
            circuit.x(3).c_if(cr, 10)

        expected = "\n".join(
            [
                "                  ",
                " q_2: ────────────",
                "      ┌───┐ ┌───┐ ",
                " q_3: ┤ H ├─┤ X ├─",
                "      ├───┤ └─╥─┘ ",
                " q_0: ┤ H ├───╫───",
                "      ├───┤   ║   ",
                " q_1: ┤ X ├───╫───",
                "      └───┘┌──╨──┐",
                " c: 4/═════╡ 0xa ╞",
                "           └─────┘",
                "ca: 2/════════════",
                "                  ",
            ]
        )
        result = visualization.circuit_drawer(
            circuit, output="text", wire_order=[2, 3, 0, 1], cregbundle=True
        )
        self.assertEqual(result.__str__(), expected)

    def test_wire_order_raises(self):
        """Verify we raise if using wire order incorrectly."""

        circuit = QuantumCircuit(3, 3)
        circuit.x(1)

        with self.assertRaisesRegex(VisualizationError, "should not have repeated elements"):
            visualization.circuit_drawer(circuit, wire_order=[2, 1, 0, 3, 1, 5])

        with self.assertRaisesRegex(VisualizationError, "cannot be set when the reverse_bits"):
            visualization.circuit_drawer(circuit, wire_order=[0, 1, 2, 5, 4, 3], reverse_bits=True)

        with self.assertWarnsRegex(RuntimeWarning, "cregbundle set"):
            visualization.circuit_drawer(circuit, cregbundle=True, wire_order=[0, 1, 2, 5, 4, 3])

    def test_reverse_bits(self):
        """Test reverse_bits should not raise warnings when no classical qubits:
        See: https://github.com/Qiskit/qiskit-terra/pull/8689"""
        circuit = QuantumCircuit(3)
        circuit.x(1)
        expected = "\n".join(
            [
                "          ",
                "q_2: ─────",
                "     ┌───┐",
                "q_1: ┤ X ├",
                "     └───┘",
                "q_0: ─────",
                "          ",
            ]
        )
        result = visualization.circuit_drawer(circuit, output="text", reverse_bits=True)
        self.assertEqual(result.__str__(), expected)

    def test_warning_for_bad_justify_argument(self):
        """Test that the correct DeprecationWarning is raised when the justify parameter is badly input,
        for both of the public interfaces."""
        circuit = QuantumCircuit()
        bad_arg = "bad"
        error_message = re.escape(
            f"Setting QuantumCircuit.draw()’s or circuit_drawer()'s justify argument: {bad_arg}, to a "
            "value other than 'left', 'right', 'none' or None (='left'). Default 'left' will be used. "
            "Support for invalid justify arguments is deprecated as of Qiskit 1.2.0. Starting no "
            "earlier than 3 months after the release date, invalid arguments will error.",
        )

        with self.assertWarnsRegex(DeprecationWarning, error_message):
            visualization.circuit_drawer(circuit, justify=bad_arg)

        with self.assertWarnsRegex(DeprecationWarning, error_message):
            circuit.draw(justify=bad_arg)

    @unittest.skipUnless(optionals.HAS_PYLATEX, "needs pylatexenc for LaTeX conversion")
    def test_no_explict_cregbundle(self):
        """Test no explicit cregbundle should not raise warnings about being disabled
        See: https://github.com/Qiskit/qiskit-terra/issues/8690"""
        inner = QuantumCircuit(1, 1, name="inner")
        inner.measure(0, 0)
        circuit = QuantumCircuit(2, 2)
        circuit.append(inner, [0], [0])
        expected = "\n".join(
            [
                "     ┌────────┐",
                "q_0: ┤0       ├",
                "     │        │",
                "q_1: ┤  inner ├",
                "     │        │",
                "c_0: ╡0       ╞",
                "     └────────┘",
                "c_1: ══════════",
                "               ",
            ]
        )
        result = circuit.draw("text")
        self.assertEqual(result.__str__(), expected)
        # Extra tests that no cregbundle (or any other) warning is raised with the default settings
        # for the other drawers, if they're available to test.
        circuit.draw("latex_source")
        if optionals.HAS_MATPLOTLIB and optionals.HAS_PYLATEX:
            circuit.draw("mpl")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

# Since the import is nearly entirely delegated to an external package, most of the testing is done
# there.  Here we need to test our wrapping behavior for base functionality and exceptions.  We
# don't want to get into a situation where updates to `qiskit_qasm3_import` breaks Terra's test
# suite due to too specific tests on the Terra side.

import os
import tempfile
import unittest
import warnings

from qiskit import qasm3
from qiskit.exceptions import ExperimentalWarning
from qiskit.circuit import QuantumCircuit, QuantumRegister, ClassicalRegister, Qubit, Clbit
from qiskit.circuit import library as lib
from qiskit.utils import optionals
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@unittest.skipUnless(optionals.HAS_QASM3_IMPORT, "need qiskit-qasm3-import for OpenQASM 3 imports")
class TestOldQASM3Import(QiskitTestCase):
    # These tests are for the `qiskit-qasm3-import` hooks, not the native one.

    def test_import_errors_converted(self):
        with self.assertRaises(qasm3.QASM3ImporterError):
            qasm3.loads("OPENQASM 3.0; qubit[2.5] q;")

    def test_loads_can_succeed(self):
        program = """
            OPENQASM 3.0;
            include "stdgates.inc";
            qubit[2] qr;
            bit[2] cr;
            h qr[0];
            cx qr[0], qr[1];
            cr[0] = measure qr[0];
            cr[1] = measure qr[1];
        """
        parsed = qasm3.loads(program)
        expected = QuantumCircuit(QuantumRegister(2, "qr"), ClassicalRegister(2, "cr"))
        expected.h(0)
        expected.cx(0, 1)
        expected.measure(0, 0)
        expected.measure(1, 1)
        self.assertEqual(parsed, expected)

    def test_load_can_succeed(self):
        program = """
            OPENQASM 3.0;
            include "stdgates.inc";
            qubit[2] qr;
            bit[2] cr;
            h qr[0];
            cx qr[0], qr[1];
            cr[0] = measure qr[0];
            cr[1] = measure qr[1];
        """
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = os.path.join(tmp_dir, "bell.qasm")
            with open(tmp_path, "w") as fptr:
                fptr.write(program)
            parsed = qasm3.load(tmp_path)
        expected = QuantumCircuit(QuantumRegister(2, "qr"), ClassicalRegister(2, "cr"))
        expected.h(0)
        expected.cx(0, 1)
        expected.measure(0, 0)
        expected.measure(1, 1)
        self.assertEqual(parsed, expected)


class TestQASM3Import(QiskitTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._cm = warnings.catch_warnings()
        cls._cm.__enter__()
        # We're knowingly testing the experimental code.
        warnings.filterwarnings("ignore", category=ExperimentalWarning, module="qiskit.qasm3")

    @classmethod
    def tearDownClass(cls):
        cls._cm.__exit__(None, None, None)
        super().tearDownClass()

    def test_load_can_succeed(self):
        """Basic test of `load` - everything else we'll do via `loads` because it's easier."""
        program = """
            OPENQASM 3.0;
            include "stdgates.inc";
            qubit[2] qr;
            bit[2] cr;
            h qr[0];
            cx qr[0], qr[1];
            cr[0] = measure qr[0];
            cr[1] = measure qr[1];
        """
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = os.path.join(tmp_dir, "bell.qasm")
            with open(tmp_path, "w") as fptr:
                fptr.write(program)
            parsed = qasm3.load_experimental(tmp_path)
        expected = QuantumCircuit(QuantumRegister(2, "qr"), ClassicalRegister(2, "cr"))
        expected.h(0)
        expected.cx(0, 1)
        expected.measure(0, 0)
        expected.measure(1, 1)
        self.assertEqual(parsed, expected)

    def test_simple_loose_bits(self):
        program = """
            OPENQASM 3.0;
            include "stdgates.inc";
            qubit q0;
            qubit q1;
            bit c0;
            bit c1;

            h q0;
            cx q0, q1;
            c0 = measure q0;
            c1 = measure q1;
        """
        parsed = qasm3.loads_experimental(program)
        expected = QuantumCircuit([Qubit(), Qubit(), Clbit(), Clbit()])
        expected.h(0)
        expected.cx(0, 1)
        expected.measure(0, 0)
        expected.measure(1, 1)
        self.assertEqual(parsed, expected)

    def test_all_stdlib_gates(self):
        # Notably this doesn't include negative floating-point values yet.  Can be fixed after:
        #   https://github.com/Qiskit/openqasm3_parser/issues/81
        program = """
            OPENQASM 3.0;
            include "stdgates.inc";
            qubit[3] q;
            p(0.5) q[0];
            x q[0];
            y q[0];
            z q[0];
            h q[0];
            s q[0];
            sdg q[0];
            t q[0];
            tdg q[0];
            sx q[0];
            rx(0.5) q[0];
            ry(0.5) q[0];
            rz(0.5) q[0];
            cx q[0], q[1];
            cy q[0], q[1];
            cz q[0], q[1];
            cp(1.5) q[0], q[1];
            crx(0.25) q[0], q[1];
            cry(0.75) q[0], q[1];
            crz(0.5) q[0], q[1];
            ch q[0], q[1];
            swap q[0], q[1];
            ccx q[0], q[1], q[2];
            cswap q[0], q[1], q[2];
            cu(0.25, 0.5, 0.75, 1.0) q[0], q[1];
            CX q[0], q[1];
            phase(0.5) q[0];
            cphase(0.5) q[0], q[1];
            id q[0];
            u1(0.5) q[0];
            u2(0.25, 0.5) q[0];
            u3(0.25, 0.5, 0.75) q[0];
        """
        parsed = qasm3.loads_experimental(program)
        expected = QuantumCircuit(QuantumRegister(3, "q"))
        expected.p(0.5, 0)
        expected.x(0)
        expected.y(0)
        expected.z(0)
        expected.h(0)
        expected.s(0)
        expected.sdg(0)
        expected.t(0)
        expected.tdg(0)
        expected.sx(0)
        expected.rx(0.5, 0)
        expected.ry(0.5, 0)
        expected.rz(0.5, 0)
        expected.cx(0, 1)
        expected.cy(0, 1)
        expected.cz(0, 1)
        expected.cp(1.5, 0, 1)
        expected.crx(0.25, 0, 1)
        expected.cry(0.75, 0, 1)
        expected.crz(0.5, 0, 1)
        expected.ch(0, 1)
        expected.swap(0, 1)
        expected.ccx(0, 1, 2)
        expected.cswap(0, 1, 2)
        expected.cu(0.25, 0.5, 0.75, 1, 0, 1)
        expected.cx(0, 1)
        expected.p(0.5, 0)
        expected.cp(0.5, 0, 1)
        expected.id(0)
        expected.append(lib.U1Gate(0.5), [0], [])
        expected.append(lib.U2Gate(0.25, 0.5), [0], [])
        expected.append(lib.U3Gate(0.25, 0.5, 0.75), [0], [])
        self.assertEqual(parsed, expected)

    def test_barrier(self):
        program = """
            OPENQASM 3.0;
            qubit[2] a;
            qubit b;
            qubit[5] c;
            barrier b;
            barrier a;
            barrier a, b;
            barrier a[0], c[2];
            barrier c[{3, 1, 4}][{2, 0, 1}][0], b;
        """
        parsed = qasm3.loads_experimental(program)
        a, b, c = QuantumRegister(2, "a"), Qubit(), QuantumRegister(5, "c")
        expected = QuantumCircuit(a, [b], c)
        expected.barrier(b)
        expected.barrier(a)
        expected.barrier(a, b)
        expected.barrier(a[0], c[2])
        expected.barrier(c[4], b)
        self.assertEqual(parsed, expected)

    def test_measure(self):
        program = """
            OPENQASM 3.0;
            qubit[2] q0;
            qubit q1;
            qubit[5] q2;
            bit[2] c0;
            bit c1;
            bit[5] c2;
            c0 = measure q0;
            c0[0] = measure q0[0];
            c1 = measure q1;
            c2[{3, 1, 4}] = measure q2[{4, 2, 3}];
        """
        parsed = qasm3.loads_experimental(program)
        q0, q1, q2 = QuantumRegister(2, "q0"), Qubit(), QuantumRegister(5, "q2")
        c0, c1, c2 = ClassicalRegister(2, "c0"), Clbit(), ClassicalRegister(5, "c2")
        expected = QuantumCircuit(q0, [q1], q2, c0, [c1], c2)
        expected.measure(q0, c0)
        expected.measure(q0[0], c0[0])
        expected.measure(q1, c1)
        expected.measure(q2[[4, 2, 3]], c2[[3, 1, 4]])
        self.assertEqual(parsed, expected)

    def test_override_custom_gate_parameterless(self):
        program = """
            OPENQASM 3.0;
            gate my_gate a, b {}
            qubit[2] q;
            my_gate q[0], q[1];
            my_gate q[1], q[0];
        """
        parsed = qasm3.loads_experimental(
            program, custom_gates=[qasm3.CustomGate(lib.CXGate, "my_gate", 0, 2)]
        )
        expected = QuantumCircuit(QuantumRegister(2, "q"))
        expected.cx(0, 1)
        expected.cx(1, 0)
        self.assertEqual(parsed, expected)

    def test_override_custom_gate_parametric(self):
        program = """
            OPENQASM 3.0;
            gate my_crx(ang) a, b {}
            qubit[2] q;
            my_crx(0.5) q[0], q[1];
            my_crx(1.5) q[1], q[0];
        """
        parsed = qasm3.loads_experimental(
            program, custom_gates=[qasm3.CustomGate(lib.CRXGate, "my_crx", 1, 2)]
        )
        expected = QuantumCircuit(QuantumRegister(2, "q"))
        expected.crx(0.5, 0, 1)
        expected.crx(1.5, 1, 0)
        self.assertEqual(parsed, expected)

    def test_set_include_path(self):
        include = """
            gate my_gate a {}
        """
        program = """
            OPENQASM 3.0;
            include "my_include.qasm";
            qubit q0;
            my_gate q0;
        """
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = os.path.join(tmp_dir, "my_include.qasm")
            with open(tmp_path, "w") as fptr:
                fptr.write(include)
            # Can't test for failed import yet due to:
            #   https://github.com/Qiskit/openqasm3_parser/issues/74
            parsed = qasm3.loads_experimental(
                program,
                custom_gates=[qasm3.CustomGate(lib.XGate, "my_gate", 0, 1)],
                include_path=[tmp_dir],
            )
        expected = QuantumCircuit([Qubit()])
        expected.x(0)
        self.assertEqual(parsed, expected)

    def test_gate_broadcast(self):
        program = """
            OPENQASM 3.0;
            include "stdgates.inc";
            qubit[2] q0;
            qubit q1;
            qubit[2] q2;
            h q0;
            cx q0, q1;
            cx q0[0], q2;
            cx q0, q2;
            ccx q0[{1, 0}], q1, q2;
        """
        parsed = qasm3.loads_experimental(program)
        q0, q1, q2 = QuantumRegister(2, "q0"), Qubit(), QuantumRegister(2, "q2")
        expected = QuantumCircuit(q0, [q1], q2)
        expected.h(q0[0])
        expected.h(q0[1])
        #
        expected.cx(q0[0], q1)
        expected.cx(q0[1], q1)
        #
        expected.cx(q0[0], q2[0])
        expected.cx(q0[0], q2[1])
        #
        expected.cx(q0[0], q2[0])
        expected.cx(q0[1], q2[1])
        #
        expected.ccx(q0[1], q1, q2[0])
        expected.ccx(q0[0], q1, q2[1])
        self.assertEqual(parsed, expected)

    def test_custom_gate_inspectable(self):
        """Test that the `CustomGate` object can be inspected programmatically after creation."""
        custom = qasm3.CustomGate(lib.CXGate, "cx", 0, 2)
        self.assertEqual(custom.name, "cx")
        self.assertEqual(custom.num_params, 0)
        self.assertEqual(custom.num_qubits, 2)

        self.assertIsInstance(qasm3.STDGATES_INC_GATES[0], qasm3.CustomGate)
        stdgates = {
            gate.name: (gate.num_params, gate.num_qubits) for gate in qasm3.STDGATES_INC_GATES
        }
        self.assertEqual(stdgates["rx"], (1, 1))
        self.assertEqual(stdgates["cphase"], (1, 2))



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Primitive V1 abstract base class."""

from __future__ import annotations

from abc import ABC

from qiskit.providers import Options


class BasePrimitive(ABC):
    """Primitive V1 abstract base class."""

    def __init__(self, options: dict | None = None):
        self._run_options = Options()
        if options is not None:
            self._run_options.update_options(**options)

    @property
    def options(self) -> Options:
        """Return options values for the estimator.

        Returns:
            options
        """
        return self._run_options

    def set_options(self, **fields):
        """Set options values for the estimator.

        Args:
            **fields: The fields to update the options
        """
        self._run_options.update_options(**fields)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Mixin for gate operator interface.
"""

import sys
from abc import ABC, abstractmethod
from numbers import Integral

from qiskit.exceptions import QiskitError

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class GroupMixin(ABC):
    """Abstract Mixin for operator group operations.

    This class defines the following methods

        - :meth:`compose`
        - :meth:`dot`
        - :meth:`tensor`
        - :meth:`expand`
        - :meth:`power`

    And the following operator overloads:

        - ``&``, ``__and__`` -> :meth:`compose`
        - ``@``, ``__matmul__`` -> :meth:`dot`
        - ``^``, ``__xor__`` -> `:meth:`tensor`
        - ``**``, ``__pow__`` -> :meth:`power`

    The following abstract methods must be implemented by subclasses
    using this mixin

        - ``compose(self, other, qargs=None, inplace=False)``
        - ``tensor(self, other)``
        - ``expand(self, other)``
    """

    def __and__(self, other) -> Self:
        return self.compose(other)

    def __pow__(self, n) -> Self:
        return self.power(n)

    def __xor__(self, other) -> Self:
        return self.tensor(other)

    def __matmul__(self, other) -> Self:
        return self.dot(other)

    @abstractmethod
    def tensor(self, other) -> Self:
        r"""Return the tensor product with another CLASS.

        Args:
            other (CLASS): a CLASS object.

        Returns:
            CLASS: the tensor product :math:`a \otimes b`, where :math:`a`
                is the current CLASS, and :math:`b` is the other CLASS.

        .. note::
            The tensor product can be obtained using the ``^`` binary operator.
            Hence ``a.tensor(b)`` is equivalent to ``a ^ b``.

        .. note:
            Tensor uses reversed operator ordering to :meth:`expand`.
            For two operators of the same type ``a.tensor(b) = b.expand(a)``.
        """

    @abstractmethod
    def expand(self, other) -> Self:
        r"""Return the reverse-order tensor product with another CLASS.

        Args:
            other (CLASS): a CLASS object.

        Returns:
            CLASS: the tensor product :math:`b \otimes a`, where :math:`a`
                is the current CLASS, and :math:`b` is the other CLASS.

        .. note:
            Expand is the opposite operator ordering to :meth:`tensor`.
            For two operators of the same type ``a.expand(b) = b.tensor(a)``.
        """

    @abstractmethod
    def compose(self, other, qargs=None, front=False) -> Self:
        """Return the operator composition with another CLASS.

        Args:
            other (CLASS): a CLASS object.
            qargs (list or None): Optional, a list of subsystem positions to
                                  apply other on. If None apply on all
                                  subsystems (default: None).
            front (bool): If True compose using right operator multiplication,
                          instead of left multiplication [default: False].

        Returns:
            CLASS: The composed CLASS.

        Raises:
            QiskitError: if other cannot be converted to an operator, or has
                         incompatible dimensions for specified subsystems.

        .. note::
            Composition (``&``) by default is defined as `left` matrix multiplication for
            matrix operators, while ``@`` (equivalent to :meth:`dot`) is defined as `right` matrix
            multiplication. That is that ``A & B == A.compose(B)`` is equivalent to
            ``B @ A == B.dot(A)`` when ``A`` and ``B`` are of the same type.

            Setting the ``front=True`` kwarg changes this to `right` matrix
            multiplication and is equivalent to the :meth:`dot` method
            ``A.dot(B) == A.compose(B, front=True)``.
        """

    def dot(self, other, qargs=None) -> Self:
        """Return the right multiplied operator self * other.

        Args:
            other (CLASS): an operator object.
            qargs (list or None): Optional, a list of subsystem positions to
                                  apply other on. If None apply on all
                                  subsystems (default: None).

        Returns:
            CLASS: The right matrix multiplied CLASS.

        .. note::
            The dot product can be obtained using the ``@`` binary operator.
            Hence ``a.dot(b)`` is equivalent to ``a @ b``.
        """
        return self.compose(other, qargs=qargs, front=True)

    def power(self, n) -> Self:
        """Return the compose of a operator with itself n times.

        Args:
            n (int): the number of times to compose with self (n>0).

        Returns:
            CLASS: the n-times composed operator.

        Raises:
            QiskitError: if the input and output dimensions of the operator
                         are not equal, or the power is not a positive integer.
        """
        # NOTE: if a subclass can have negative or non-integer powers
        # this method should be overridden in that class.
        if not isinstance(n, Integral) or n < 1:
            raise QiskitError("Can only power with positive integer powers.")
        ret = self
        for _ in range(1, n):
            ret = ret.dot(self)
        return ret
